#!/bin/python
# MISTIC Project INRIA/INRAE
# Author Muller Coralie
# Date: 2024/11/27
# Update: 2025/08/-

####################################
#            MAPPING               #
####################################

# Script to match metabolites between MAF files and SBML files.
# A conversion table is used to perform the matching. This table is
# based on the MetaCyc/MetaNetX database, and includes corresponding
# information and IDs from other databases to establish links.


import ast
import asyncio
import csv
import itertools
import re
import sys
import time
from pathlib import Path

import cobra
import pandas as pd

from metanetmap import utils

####################################
#          Main Script             #
####################################

logger = utils.get_logger("Mapping")


# ---------------------------------#
#     Load the conversion table   #
# ---------------------------------#


def load_database(database_conversion):
    """
    Load the database_conversion as a dictionary
    /!\\ Special handling is applied to the 'SYNONYMS' column: if it's
    a non-empty string, it is converted from a string representation
    of a list to an actual Python list.

    Exemple of the structure :
          [{'UNIQUE-ID': 'CPD-17659', 'CHEBI': '85306', 'COMMON-NAME':
          'D-allo-isoleucine', 'ABBREV-NAME': '',
          'SYNONYMS': [], 'ADD-COMPLEMENT': '', 'SEED': '', 'BIGG':
          '', 'HMDB': '', 'METANETX': 'MNXM17053', 'LIGAND-CPD':
          'C21092',
          'REFMET': '', 'PUBCHEM': '6950184', 'CAS': '', 'INCHI-KEY':
          'InChIKey=AGPKZVBTJJNPAG-CRCLSJGQSA-N', 'SMILES':
          'CC[C@H](C)[C@@H]([NH3+])C([O-])=O'},{...}]

    Args:
        database_conversion (path): Path to the database_conversion to
        exhange into dict

    Returns:
        datatable (list): List with all the dictionnaries with
        database_conversion information -> one dic corresponding to
        one row
    """
    datatable = []
    path = Path(database_conversion)

    # Verify that the database_conversion file has been properly loaded
    try:
        with path.open(encoding="utf-8") as f:
            reader = csv.DictReader(f, delimiter="\t")
            for row in reader:
                if "SYNONYMS" in row and row["SYNONYMS"]:
                    try:
                        row["SYNONYMS"] = ast.literal_eval(row["SYNONYMS"])
                    except Exception:
                        logger.warning(f"Failed to parse SYNONYMS: {row['SYNONYMS']}")
                        row["SYNONYMS"] = []
                    
                # Format CHEBI
                chebi = row.get("CHEBI")
                if chebi:
                    row["CHEBI"] = f"CHEBI:{chebi}"
                # Format PUBCHEM
                pubchem = row.get("PUBCHEM")
                if pubchem:
                    row["PUBCHEM"] = f"PUBCHEM:{pubchem}"
                datatable.append(row)
    except Exception as e:
        logger.critical(f"Error reading file: {e}")
        sys.exit(1)

    return datatable


# ---------------------------------------------------------------------#
#  Preparation sbml and maf list with informations for the mapping    #
# ---------------------------------------------------------------------#


# Set Paths
def set_list_paths(paths_to_list, list_pathways, ext1, ext2):
    """
    Create a list of file paths from a directory or a single file,
    filtering by extension if provided.

    Args:
        paths_to_list (Path | str): Path to a directory or single file
          containing pathway input files.
        list_pathways (List[str]): List to store valid file paths.
        ext1 (str | None): First allowed file extension (e.g., ".sbml").
        ext2 (str | None): Second allowed file extension (e.g., ".xml").

    Returns:
        list_pathways (List[str]): Updated list containing valid file
        paths.
    """
    paths_to_list = Path(paths_to_list)

    if paths_to_list.is_dir():  # If it's a directory
        for path_file in paths_to_list.iterdir():
            if path_file.is_file():
                if ext1:
                    if path_file.suffix == ext1 or path_file.suffix == ext2:
                        list_pathways.append(str(path_file))
                    else:
                        logger.critical(
                            "/!\\ Problem with extension of the file "
                            f"{path_file}: must be {ext1} or {ext2}"
                        )
                        sys.exit()
                else:
                    list_pathways.append(str(path_file))

    elif paths_to_list.is_file():  # If it's a single file
        if ext1:
            if paths_to_list.suffix == ext1 or paths_to_list.suffix == ext2:
                list_pathways.append(str(paths_to_list))
            else:
                logger.critical(
                    "/!\\ Problem with extension of the file "
                    f"{paths_to_list}: must be {ext1} or {ext2}"
                )
                sys.exit()
        else:
            list_pathways.append(str(paths_to_list))

    else:
        logger.critical(
            f"/!\\ The provided path {paths_to_list} is neither a file nor a directory."
        )
        sys.exit()

    return list_pathways


def manage_id_in_metadata_sbml(annotations, tmp_data):
    """
    Cleans and extracts IDs from the SBML annotations and stores them
    in tmp_data.

    This function processes each key-value pair in the `annotations`
    dictionary:
    - If the value is a list, it removes common database prefixes
      (e.g., "CHEBI:", "META:") from each element.
    - If the value is a single string, it removes the prefixes and
      stores it as a list with one element.

    The cleaned results are stored in the `tmp_data` dictionary under
    the same keys.

    Args:
        annotations (dict): Dictionary of annotations (typically from
          SBML metadata). Keys are database names, values can be strings
          or lists of strings.
        tmp_data (dict): Temporary dictionary where cleaned IDs are
          stored.

    Returns:
        tmp_data (dict): Updated tmp_data with cleaned annotation
        values.
    """
    for key, value in annotations.items():
        # Special case: inchikey must always be a string, not a list
        if key == "inchikey":
            tmp_data[key] = [f"INCHIKEY={str(value).strip()}"]
            continue

        if key == "pubchem.compound":
            tmp_data[key] = [f"PUBCHEM:{str(value).strip()}"]
            continue

        tmp_data[key] = []
        if isinstance(value, list):
            tmp_data[key] = [
                str(i).replace("META:", "") for i in value
            ]
        else:
            value_clean = str(value).replace("META:", "")
            tmp_data[key].append(value_clean)
    return tmp_data


def merge_doublons_metadata_sbml(meta_data_sbml, met_name, tmp_data):
    """
    Merge new annotation data (`tmp_data`) into the main SBML metadata
    dictionary (`meta_data_sbml`), avoiding duplicates.

    For each key in `tmp_data` (e.g., database name), the function
    checks if the corresponding values already exist in
    `meta_data_sbml[met_name]`, and if not, appends them.

    Args:
        meta_data_sbml (dict): Main dictionary of metadata for all
          metabolites. Structure: {metabolite_name: {source_db: [ids]}}
        met_name (str): The metabolite name (key in `meta_data_sbml`)
        to update.
        tmp_data (dict): Temporary dictionary containing new metadata
        to merge.

    Returns:
        meta_data_sbml (dict): Updated `meta_data_sbml` with new
        entries added without duplicates.
    """
    for k, v in tmp_data.items():
        # If meta_data_sbml[met_name] doesn't contain this key k yet, create it with an empty list []
        meta_data_sbml[met_name].setdefault(k, [])
        for i in v:
            # Flatten all values in meta_data_sbml[met_name] and check
            # if `i` already exists.
            if i not in itertools.chain.from_iterable(meta_data_sbml[met_name].values()):
                meta_data_sbml[met_name][k].append(i)
        # Remove duplicates after the loop (more efficient)
        meta_data_sbml[met_name][k] = list(dict.fromkeys(meta_data_sbml[met_name][k]))
    return meta_data_sbml



def extract_metadata_sbml(model, meta_data_sbml):
    """
    Extracts annotations and metadata for each metabolite in an SBML
    model and merges the data into the meta_data_sbml dictionary.

    For each metabolite:
    - Adds its ID and formula
    - Parses annotation fields (CHEBI, HMDB, etc.)
    - Avoids duplicates when merging data

    Args:
        model (Model): The SBML model containing metabolites (from
          COBRApy).
        meta_data_sbml (dict): Dictionary where metadata is collected
          and updated. Structure: {metabolite_name: {annotation_type: [values]}}

    Returns:
        meta_data_sbml (dict): Updated meta_data_sbml with metabolite
        metadata from the SBML model.
    """
    # Load and parse the SBML file
    for m in model.metabolites:
        tmp_data = {}
        annotations = m.annotation
        # Existing entry: update ID and formula
        if m.name in meta_data_sbml.keys():
            if m.id.endswith(']'):
                clean_id=m.id.rsplit("[", 1)[0]
            else:
                clean_id=m.id.rsplit("_", 1)[0]
            meta_data_sbml[m.name]["ID"].append(clean_id)
            if m.formula:
                meta_data_sbml[m.name]["formula"].append(m.formula)
            else:
                tmp_data["formula"] = ""
            # Remove duplicates from ID and formula
            for key in ["ID", "formula"]:
                meta_data_sbml[m.name][key] = list(
                    dict.fromkeys(meta_data_sbml[m.name][key])
                )
            # Extract other metadata and merge
            tmp_data = manage_id_in_metadata_sbml(annotations, tmp_data)
            meta_data_sbml = merge_doublons_metadata_sbml(
                meta_data_sbml, m.name, tmp_data
            )
        else:
            # New entry
            tmp_data = {k: [] for k in annotations.keys()}
            tmp_data["ID"] = []
            tmp_data["NAME"]=[m.name]
            if m.id.endswith(']'):
                tmp_data["ID"].append(m.id.rsplit("[", 1)[0])
            else:
                tmp_data["ID"].append(m.id.rsplit("_", 1)[0])
            if m.formula:
                tmp_data["formula"] = [m.formula]
            else:
                tmp_data["formula"] = ""
            tmp_data = manage_id_in_metadata_sbml(annotations, tmp_data)
            meta_data_sbml[m.name] = tmp_data
    return meta_data_sbml


# SBML
def setup_merge_list_sbml_metabolites(List_SBML_paths):
    """
    Creates a list with all IDs of metabolites in all metabolic
    networks (for the community mode) and a directory with key: name
    of one metabolic networks , value the list of metabolites for this
    metabolic networks.

    Args:
        List_SBML_paths (List): List with paths for metabolic networks
          files

    Returns:
        dic_couple_sbml (dic): Dictionary with:  key-> name of one
          metabolic networks , value -> list of metabolites for this
          sbml
        meta_data_sbml (dict): Updated meta_data_sbml with metabolite
          metadata from the SBML model.
    """
    dic_couple_sbml = {}
    meta_data_sbml_merge = {}
    list_metabolites_id_merge = []

    for path_sbml in List_SBML_paths:
        model = cobra.io.read_sbml_model(path_sbml)

        # Extract and merge metadata from the ALL modelS
        meta_data_sbml_merge = extract_metadata_sbml(model, meta_data_sbml_merge)

        # Extract and merge metadata from ONE UNQUE model
        meta_data_sbml = extract_metadata_sbml(model, {})
        flat_list = [ value for compound in meta_data_sbml.values() for v in compound.values()  for value in v]

        # Remove compartment suffix (e.g., "_c", "_e") from metabolite IDs
        list_metabolites_id=[]

        for m in model.metabolites:
            if m.id.endswith(']'):
                clean_id=m.id.rsplit("[", 1)[0]
            else:
                clean_id=m.id.rsplit("_", 1)[0]
            list_metabolites_id.append(clean_id)

        # Add to global merged list
        list_metabolites_id_merge.append(list_metabolites_id)

        # Store metabolite list for this specific SBML file
        dic_couple_sbml[Path(path_sbml).stem] = list_metabolites_id+flat_list

    # Flatten and deduplicate global list
    list_metabolites_id_merge = list(
        dict.fromkeys(itertools.chain.from_iterable(list_metabolites_id_merge))
    )
    return dic_couple_sbml, meta_data_sbml_merge


# Metabolomic files
def setup_merged_list_maf_metabolites(List_MAF_paths):
    """
    Parses a list of MAF (Metabolite Annotation File) files to extract
    and clean metadata, and merges them into a single dictionary and
    DataFrame for downstream comparison.

    Args:
        List_MAF_paths (List[str]): List of file paths to MAF files
        (TSV format).

    Returns:
        maf_dictionnary_clean (Dict[str, List[str]]): Cleaned
          dictionary with keys as column names and values as unique,
          non-empty data.
        keys (List[str]): Original list of keys expected/allowed in
          MAF files.
        maf_merged_df (pd.DataFrame): Merged and deduplicated
          DataFrame from all input MAF files.
    """
    maf_dictionnary_clean = {}
    df_list = []

    # Prepare the maf dictionnary
    maf_dictionnary = {}
    keys = [
        "UNIQUE-ID",
        "CHEBI",
        "COMMON-NAME",
        "ABBREV-NAME",
        "SYNONYMS",
        "ADD-COMPLEMENT",
        "MOLECULAR-WEIGHT",
        "MONOISOTOPIC-MW",
        "SEED",
        "BIGG",
        "HMDB",
        "METANETX",
        "METACYC",
        "LIGAND-CPD",
        "REFMET",
        "PUBCHEM",
        "VMH",
        "CAS",
        "INCHI",
        "NON-STANDARD-INCHI",
        "INCHI-KEY",
        "SMILES",
        "FORMULA",
    ] 

    # Initialize dictionary with empty lists for each key
    for k in keys:
        maf_dictionnary[k] = []

    # Loop over MAF file paths
    for path_maf in List_MAF_paths:
        try:
            df = pd.read_csv(path_maf, sep="\t", header=0)
        except Exception as e:
            logger.critical(f"/!\\ Failed to read MAF file {path_maf}: {e}")
            sys.exit(1)
        # Read and concatenate all files
        df.columns = df.columns.str.upper()
        df_list.append(df)
        
        ##### Check if at least on column name fit with maf file
        # Convert the column index (which is a pandas.Index object) into a standard Python list
        column_names_check = list(df.columns)
        # Find matches between actual column names and expected names
        matches = set(column_names_check) & set(keys)

        # If there are no matches, raise an error
        if not matches:
            logger.critical( f"None of the expected names match the column names in the file. "
                            f"Check the following allowed column names: {keys}")
            sys.exit(1)


        # Check for tab separation issue (often due to CSV or Excel
        # file saved as wrong format).
        if df.shape[1] == 1:
            logger.critical(
                f"/!\\ The format of '{path_maf}' is not valid. "
                "Make sure it's tab-delimited (TSV)."
            )
            sys.exit(1)

        for col in df.columns:  # Loop on each column of the maf file
            col_upper = col.upper()
            if col_upper in keys:
                df[col] = df[col].astype(str).str.strip()

                # Fix floats formatted as strings with trailing '.0' (e.g., '123456.0' → '123456')
                if col_upper in ["CHEBI", "PUBCHEM"]:
                    df[col] = df[col].apply(
                        lambda x: x[:-2] if isinstance(x, str) and x.endswith(".0") else x
                    )


                # Normalize identifiers for CHEBI column
                if col_upper == "CHEBI":
                    df[col] = df[col].apply(lambda x: (
                        f"CHEBI:{x.split(':', 1)[1].strip()}" if x.lower().startswith("chebi:") else (
                            f"CHEBI:{x}" if x.isdigit() else x
                        )
                    ))

                # Normalize identifiers for PUBCHEM column
                elif col_upper == "PUBCHEM":
                    df[col] = df[col].apply(lambda x: (
                        f"PUBCHEM:{x.split(':', 1)[1].strip()}" if x.lower().startswith("pubchem:") else (
                            f"PUBCHEM:{x}" if x.isdigit() else x
                        )
                    ))
                
                # Normalize identifiers for INCHI-KEY column
                elif col_upper == "INCHI-KEY":
                    df[col] = df[col].apply(lambda x: (
                        f"INCHIKEY={x.split('=', 1)[1].strip()}" if isinstance(x, str) and x.lower().startswith("inchikey=")
                        else f"INCHIKEY={x.strip()}" if isinstance(x, str) and x.strip().lower() not in ["", "nan"]
                        else x
                    ))

                # Common cleaning for all columns
                df[col] = df[col].apply(utils.fix_arrows_in_parentheses) 
                # Add to dictionary
                maf_dictionnary[col_upper].extend(df[col].tolist())  
                # Remove duplicates while preserving order
                maf_dictionnary[col_upper] = list(dict.fromkeys(maf_dictionnary[col_upper]))


    # Remove keys with only empty values (e.g., "", nan)
    maf_dictionnary_clean = utils.remove_empty_keys(maf_dictionnary)
    maf_dictionnary_clean = {
        k: v for k, v in maf_dictionnary_clean.items() if v != ["nan"]
    }

    # Merge all DataFrames and remove duplicates
    merged_df = pd.concat(df_list, ignore_index=True, sort=False)

    # Remove duplicates
    maf_merged_df = merged_df.drop_duplicates()
    maf_merged_df.columns = maf_merged_df.columns.str.upper()
    maf_merged_filtered_df = maf_merged_df[maf_dictionnary_clean.keys()]
    return maf_dictionnary_clean, keys, maf_merged_filtered_df


# --------------------------------------------------------------#
#           Manage to remove enantiomeres and Inchey           #
# --------------------------------------------------------------#


def remove_enantiomer_process(value, enantiomers=["D", "L", "R", "S"]):
    """
    Cleans enantiomer prefixes like alpha-D-, beta-D-, etc., and
    patterns like __L_ from metabolite names, preserving punctuation
    and structure, and skipping CPD identifiers.

    Args:
        value (str): The input metabolite name.
        enantiomers (list): List of enantiomer letters to clean.

    Returns:
        value.strip() (str): Cleaned name.
    """
    # Do not process identifiers
    if "CPD" in value:
        return value

    for e in enantiomers:
        # 1. Remove if at beginning (e.g., 'D-', 'alpha-D-')
        value = re.sub(
            rf"^(\()?((alpha|beta)-)?{e}-", r"\1", value, flags=re.IGNORECASE
        )

        # 2. Remove enantiomer prefixes inside parentheses (keep '(')
        value = re.sub(rf"\(((alpha|beta)-)?{e}-", "(", value, flags=re.IGNORECASE)

        # 3. Remove enantiomer prefixes AFTER commas, e.g., ', alpha-D-LKL' → ', LKL'
        value = re.sub(
            rf"(,\s*)(\()?((alpha|beta)-)?{e}-", r"\1\2", value, flags=re.IGNORECASE
        )

        # 4. Remove enantiomer prefix ONLY if followed by a closing punctuation
        value = re.sub(
            rf"((alpha|beta)-{e})(?=[\]\)\,\}}])", "", value, flags=re.IGNORECASE
        )

        # 5. Remove [alpha-D- → [ ; (beta-L- → ( ; etc.
        value = re.sub(
            rf"(?<=[\[\(\{{])((alpha|beta)-{e}-)", "", value, flags=re.IGNORECASE
        )

        # 6. Replace inner patterns like -alpha-D- or -D- by -
        value = re.sub(rf"-((alpha|beta)-)?{e}-", "-", value, flags=re.IGNORECASE)

        # 7. Fix malformed patterns like alpha--D-glucose
        value = re.sub(rf"((alpha|beta)--){e}-", "", value, flags=re.IGNORECASE)

        # 8. Remove or replace double underscore enantiomer patterns like __D_, __D
        value = re.sub(rf"__{e}(_|$)", r"\1", value, flags=re.IGNORECASE)

    return value.strip()


def remove_enantiomer_and_Inchey_db(dictionary_db):
    """
    Clean a list of dictionaries (e.g., from MetaCyc/MetaNetX) by:
    - Removing enantiomer notation ('D', 'L', 'R', 'S') from selected
      fields.
    - Trimming the InChIKey to its base fragment (first block before
    first '-').

    Args:
        dictionary_db (list of dict): List of dictionaries, each
          representing a metabolite entry.

    Returns:
        dictionary_db (list of dict): The cleaned list of dictionaries
          with modified fields.
    """
    # Columns to process
    keys_to_clean = ["UNIQUE-ID", "COMMON-NAME", "SYNONYMS", "INCHI-KEY"]

    for dico in dictionary_db:
        for key in keys_to_clean:
            value = dico.get(key)

            if not value or value in ["", [], None]:
                continue  # Skip empty values safely

            if key == "INCHI-KEY":
                # Only keep the root part before the first dash
                dico[key] = value.split("-")[0]

            else:
                # Clean enantiomer notation
                if isinstance(value, list):
                    # Clean each item in the list
                    dico[key] = [
                        remove_enantiomer_process(
                            item, enantiomers=["D", "L", "R", "S"]
                        )
                        for item in value
                    ]
                elif isinstance(value, str):
                    dico[key] = remove_enantiomer_process(
                        value, enantiomers=["D", "L", "R", "S"]
                    )

    return dictionary_db


def remove_enantiomer_and_Inchey_metadata(metadata):
    """
    Clean a dictionary of metadata for metabolites by:
    - Removing enantiomer prefixes ('D', 'L', 'R', 'S') from selected
      fields.
    - Standardizing InChIKeys to keep only the base fragment (before
      first '-').

    Args:
        metadata (dict): A dictionary where keys are metabolite names
        and values are sub-dictionaries containing metadata fields
        like 'ID', 'inchikey', 'biocyc'.

    Returns:
        metadata (dict): Updated metadata dictionary with cleaned values.
    """
    # Columns to process
    keys_to_clean = ["ID", "inchikey", "biocyc"]
    for key_meta, value_meta in metadata.items():
        for key in keys_to_clean:
            value = value_meta.get(key)
            if not value or value in ["", [], None]:
                continue  # Skip empty or invalid fields
            else:
                if isinstance(value, list):
                    # Clean each item in the list
                    value_meta[key] = [
                        remove_enantiomer_process(
                            item, enantiomers=["D", "L", "R", "S"]
                        )
                        for item in value
                    ]

    return metadata



# ------------------------------------------------------------#
#            Harmonisation output preparation                #
# ------------------------------------------------------------#


def setup_harmonisation_output(
    dic_tsv_results, keys_starter, unmatch_metabolites_total, keys
):
    """
    Add unmatched metabolites to dic_tsv_results and harmonize the
    output dictionary keys.

    Args:
        dic_tsv_results (list of dict): List of dictionaries storing
          match/unmatch results for each metabolite.
        keys_starter (list): List of base keys expected in each dict.
        unmatch_metabolites_total (list): List of metabolites
          unmatched in both databases.
        keys (list): Additional keys used to create 'Match via {key}'
          columns.

    Returns:
        dic_tsv_results (list of dict): updated list with unmatched
          metabolites added.
        keys_reorder (list): list of all keys in harmonized order.
    """
    # Add each metabolites  which unmatch in both database and
    # metaboloc network to the dic_tsv_results and add their
    # specificities (key_reorder).
    for met in unmatch_metabolites_total:
        temps = {key: (met if key == "Metabolites" else "") for key in keys_starter}
        dic_tsv_results.append(temps)

    # For each metabolite in the dic_tsv_results if a column is not in
    # dic_tsv_results, we create one with empty values.
    # Harmonize keys across all dictionaries
    keys_reorder = keys_starter.copy()  # Start with starter keys

    for dic in dic_tsv_results:
        existing_keys = set(dic.keys())

        # Add keys from keys_starter if missing
        for key in keys_starter:
            if key not in existing_keys:
                dic[key] = ""

        # Add 'Match via {key}' columns for keys in keys
        for key_in_keys in keys:
            match_key = f"Match via {key_in_keys}"
            if match_key not in dic:
                dic[match_key] = ""
            if match_key not in keys_reorder:
                keys_reorder.append(match_key)

    # Remove duplicates preserving order
    keys_reorder = list(dict.fromkeys(keys_reorder))
    return dic_tsv_results, keys_reorder


# ----------------------------------------------------#
#             Match Metabolites                      #
# ----------------------------------------------------#


def check_add_unique_id_to_sbml_match(
    sub_dict, met, Match_id, dic_tsv_results, column_name, database_info
):
    """
    Update dic_tsv_results with unique ID matches in the table
    conversion for a given metabolite.

    Args:
        sub_dict (dict): Dictionary with metabolite info from the database.
        met (str): Metabolite name.
        Match_id (dict): Dictionary tracking metabolites and their
          matched UNIQUE-ID.
        dic_tsv_results (list of dict): List of dictionaries storing
          match/unmatch results for each metabolite.
        column_name (str): Column name used for matching.
        database_info (list): List storing info about matched
          metabolites for logging.

    Returns:
        dic_tsv_results (list of dict): List of dictionaries storing
          match/unmatch results for each metabolite.
        database_info (list): List storing info about matched
          metabolites for logging.

    """
    if met in Match_id:
        sub_results_dic = utils.find_dict_by_metabolite(dic_tsv_results, met)
        if sub_results_dic:
            
            # Check if UNIQUE-ID is already present in any dic_tsv_results entry
            unique_id_exists = any(
                d.get("Match in database") == sub_dict["UNIQUE-ID"]
                for d in dic_tsv_results
            )

            if not unique_id_exists:
                
                # Get the dict where metabolite matches and update 'Match in database'
                test = next(
                    (d for d in dic_tsv_results if d.get("Metabolites") == met), None
                )

                if test and test.get("Match in database", "") == "":
                    sub_results_dic["Match in database"] = sub_dict["UNIQUE-ID"]
                elif test:
                    sub_results_dic["Match in database"] = (
                        f'{sub_dict["UNIQUE-ID"]} _AND_ {test["Match in database"]}'
                    )
                Match_id[met] = sub_dict["UNIQUE-ID"]

                # Store lowercase unique identifiers
                # from sub_dict for database_info logging
                list_value = [
                    str(value).lower() for value in sub_dict.values() if value != ""
                ]
                list_value.append(sub_dict["UNIQUE-ID"])
                database_info.append(list_value)

                logger.info(
                    f'--"{met}" is present in database with the UNIQUE-ID '
                    f'"{sub_dict["UNIQUE-ID"]}" and matches via '
                    f'"{column_name}"'
                )

            else:
                # If UNIQUE-ID already matched elsewhere, handle duplicates
                test = next(
                    (
                        d
                        for d in dic_tsv_results
                        if d.get("Match in database") == sub_dict["UNIQUE-ID"]
                    ),
                    None,
                )
                logger.info(
                    f'--"{sub_results_dic["Metabolites"]}" is a metabolite '
                    f'duplicate of "{test["Metabolites"]}" and matches via '
                    f'"{column_name}"'
                )

                # Merge metabolite names avoiding redundant concatenations
                met_names = set(sub_results_dic["Metabolites"].split(" _AND_ "))
                met_names.add(test["Metabolites"])
                sub_results_dic["Metabolites"] = " _AND_ ".join(sorted(met_names))

                sub_results_dic["Match in database"] = sub_dict["UNIQUE-ID"]
                Match_id[met] = sub_dict["UNIQUE-ID"]

                list_value = [
                    str(value).lower() for value in sub_dict.values() if value != ""
                ]
                list_value.append(sub_dict["UNIQUE-ID"])
                database_info.append(list_value)

                # Merge 'Match in metabolic networks' if exists in test
                if any(d.get("Match in database") for d in dic_tsv_results):
                    if test and test.get("Match in metabolic networks") is not None:
                        existing = sub_results_dic.get(
                            "Match in metabolic networks", []
                        )
                        # Flatten and deduplicate
                        combined = existing + (
                            test["Match in metabolic networks"]
                            if isinstance(test["Match in metabolic networks"], list)
                            else [test["Match in metabolic networks"]]
                        )
                        sub_results_dic["Match in metabolic networks"] = list(
                            set(combined)
                        )

                # Remove duplicate metabolite entry from dic_tsv_results
                dic_tsv_results = [
                    d
                    for d in dic_tsv_results
                    if d.get("Metabolites") != test["Metabolites"]
                ]
    return dic_tsv_results, database_info


def identification_sbml_match_into_db(
    dictionary_db,
    met,
    column_name,
    dic_tsv_results,
    Match_id,
    database_info,
    key_sub,
    partial,
):
    """
    Match metabolites between SBML IDs and database entries, update
    results and matched IDs.

    Args:
        dictionary_db (dict): Dictionary with metabolite info from the
          database.
        met (str): Metabolite name or partial string to search for.
        column_name (str): Column name used for matching (e.g. 'UNIQUE-ID').
        dic_tsv_results (list of dict): List of dictionaries storing
          match/unmatch results for each metabolite.
        Match_id (dict): Dictionary to track matched metabolites and
          their UNIQUE-IDs.
        database_info (list): List to store detailed match info for
          logging or output.
        key_sub (str): Key used for partial matching (e.g. metabolite
          short ID).
        partial (bool): If True, perform matching on partial keys;
          else on full metabolite name.

    Returns:
        dic_tsv_results (list of dict): List of dictionaries storing
          match/unmatch results.
        Match_id (dict): Dictionary to track matched metabolites and
          their UNIQUE-IDs.
        database_info (list): List to store detailed match info for
          logging or output.
    """
   # Find all database entries that contain the metabolite or partial string
    sub_dict = utils.find_all_entries_with_value(dictionary_db, met)

    # Use different matching keys based on partial flag ---> Partial
    # is for the last step with partial IDs.
    if partial:
        if len(sub_dict) > 1:
            # If multiple matches, iterate over all and add matches using key_sub
            for dic in sub_dict:
                dic_tsv_results, database_info = check_add_unique_id_to_sbml_match(
                    dic, key_sub, Match_id, dic_tsv_results, column_name, database_info
                )

        elif len(sub_dict) == 1:
            dic_tsv_results, database_info = check_add_unique_id_to_sbml_match(
                sub_dict[0],
                key_sub,
                Match_id,
                dic_tsv_results,
                column_name,
                database_info,
            )
    else:
        if len(sub_dict) > 1:
            # If multiple matches, iterate over all and add matches using met
            for dic in sub_dict:
                dic_tsv_results, database_info = check_add_unique_id_to_sbml_match(
                    dic, met, Match_id, dic_tsv_results, column_name, database_info
                )

        elif len(sub_dict) == 1:
            dic_tsv_results, database_info = check_add_unique_id_to_sbml_match(
                sub_dict[0], met, Match_id, dic_tsv_results, column_name, database_info
            )

    return dic_tsv_results, Match_id, database_info


def match_metab_main(
    dic, met, column_name, dic_tsv_results, Match_id, database_info, dic_temp, partial
):
    """
    Perform matching of a metabolite with database entries and update
    results dictionaries.

    Args:
        dic (dict): One entry (metabolite) from the database.
        met (str): Metabolite name from the input data.
        column_name (str): Column used for the matching
          (e.g. 'COMMON-NAME')

        dic_tsv_results (list of dict): List of dictionaries storing
          match/unmatch results for each metabolite.
        Match_id (dict): Dict linking metabolite names to their
          matched UNIQUE-IDs.
        database_info (list): List collecting detailed matching information.
        dic_temp (dict): Temporary dictionary to store new match info.
        partial (bool): Whether the match is a partial match.

    Returns:
        dic_tsv_results (list of dict): List of dictionaries storing
          match/unmatch results for each metabolite.
        database_info (list): List collecting detailed matching information.
        Match_id (dict): Dict linking metabolite names to their
          matched UNIQUE-IDs.

        if partial,
            dic_tsv_results (list of dict): List of dictionaries
              storing match/unmatch results for each metabolite.
            database_info (list): List collecting detailed matching
              information.
            Match_id (dict): Dict linking metabolite names to their
              matched UNIQUE-IDs.
            dic_temp (dict): Temporary dictionary to store new match info.
    """

    list_value = [
        str(value).lower() for value in dic.values() if value != ""
    ]  # Keep differents identifictaion for ONE metabolite

    # --- Case 1: UNIQUE-ID already matched somewhere
    if dic["UNIQUE-ID"] in Match_id.values():
        sub_results_dic = utils.find_all_entries_with_value_tsv(
            dic_tsv_results, dic["UNIQUE-ID"]
        )
        for sub_sub_results_dic in sub_results_dic:
            # Match each metabolite listed in the result dictionary
            metabolites_list = [m.strip() for m in sub_sub_results_dic["Metabolites"].split(" _AND_ ")]
            if met not in metabolites_list:
                logger.info(
                    '--"%s" is a metabolite duplicate of '
                    '"%s" and matches via "%s" ',
                    met,
                    sub_sub_results_dic["Metabolites"].split("_AND_", 2)[0].strip(),
                    column_name,
                )
                sub_sub_results_dic["Metabolites"] += f" _AND_ {met}"
                sub_sub_results_dic[f"Match via {column_name}"] = (
                    "YES"  # Add "YES" to the specific column for
                    # the doublon name which match.
                )
                Match_id[met] = dic["UNIQUE-ID"]

    # --- Case 2: New match or partial duplicate
    else:


        # Subcase 2.1: Already matched but multiple UNIQUE-IDs (partial)
        if met in Match_id.keys() and Match_id[met] != "NO UNIQUE-ID":
            
            if dic_temp:  # For metabolites with partial match AND not match in
                # STEP1 for sbml matching.
                sub_results_dic = utils.find_matching_dict(dic_tsv_results, met)
                logger.info(
                    f'--"{met}" has a partial match. We have several '
                    "UNIQUE-ID identifiers for this metabolite: "
                    f'"{sub_results_dic[f"Match in database"]}" and '
                    f'"{dic["UNIQUE-ID"]}"'
                )
                sub_results_dic["Partial match"] = (
                    f"{sub_results_dic[f'Match in database']} "
                    f"_AND_ {dic['UNIQUE-ID']}"
                )
                sub_results_dic["Match in database"] = (
                    f"{sub_results_dic[f'Match in database']} "
                    f"_AND_ {dic['UNIQUE-ID']}"
                )
                Match_id[str(met).upper()] = dic["UNIQUE-ID"]
                list_value.append(met)
                database_info.append(list_value)
            else:
                # For metabolites with partial match AND match in
                # STEP1 for sbml matching.
                sub_results_dic = utils.find_dict_by_metabolite(dic_tsv_results, met)
                if sub_results_dic:
                    logger.info(
                        f'--"{met}" has a partial match. We have '
                        "several UNIQUE-ID identifiers for this "
                        f'metabolite: "{sub_results_dic[f"Match in database"]}'
                    )
                    sub_results_dic["Partial match"] = (
                        f"{sub_results_dic[f'Match in database']} "
                        f"_AND_ {dic['UNIQUE-ID']}"
                    )

        # Subcase 2.2: Not matched yet at all
        else:
            dic_temp["Metabolites"] = met
            if partial:
                dic_temp["Partial match"] = dic["UNIQUE-ID"]
                dic_temp["Match in database"] = dic["UNIQUE-ID"]
            else:
                dic_temp["Match in database"] = dic["UNIQUE-ID"]
            Match_id[met] = dic["UNIQUE-ID"]
            dic_temp[f"Match via {column_name}"] = "YES"
            list_value.append(met)
            database_info.append(list_value)
            logger.info(
                f'--"{met}" is present in database with the UNIQUE-ID '
                f'"{dic["UNIQUE-ID"]}" and matches via "{column_name}"'
            )

    # # Add to result if new entry was created
    if dic_temp:
        dic_tsv_results.append(dic_temp)

    # Return according to matching type
    if partial:
        return dic_tsv_results, Match_id, database_info, dic_temp
    else:
        return dic_tsv_results, Match_id, database_info


def match_metabo(
    dictionary_db,
    met,
    column_name,
    dic_tsv_results,
    Match_id,
    database_info,
    key_sub,
    dic_temp_db,
    partial,
):
    """
    Match a metabolite (from SBML or MAF) to entries in the reference
    database (MetaCyc/MetaNetX).

    Args:
        dictionary_db (dict): Full MetaCyc/MetaNetX dictionary.
        met (str): Metabolite name to search.
        column_name (str): Column name used for matching (e.g. COMMON-NAME).
        dic_tsv_results (list of dict)): Output results list of dicts.
        Match_id (dict): Mapping of metabolite name -> UNIQUE-ID.
        database_info (list): List of matched metabolite data from the DB.
        key_sub (str): The original query metabolite used in partial
          matching.
        dic_temp_db (dict): Temp dict to collect partial match info.
        partial (bool): Whether to perform partial matching.

    Returns:
        dic_tsv_results (list of dict): Output results list of dicts.
        Match_id (dict): Mapping of metabolite name -> UNIQUE-ID.
        database_info (list): List of matched metabolite data from the DB.
    """
    # Find all database entries that contain the metabolite or partial string
    sub_dict = utils.find_all_entries_with_value(dictionary_db, met)
    # Use different matching keys based on partial flag ---> Partial
    # is for the last step with partial IDs.
    if partial:
        if len(sub_dict) > 1:
            # If multiple matches, iterate over all and add matches using key_sub
            for dic in sub_dict:
                dic_tsv_results, Match_id, database_info, dic_temp_db = (
                    match_metab_main(
                        dic,
                        key_sub,
                        column_name,
                        dic_tsv_results,
                        Match_id,
                        database_info,
                        dic_temp_db,
                        partial=True,
                    )
                )

        elif len(sub_dict) == 1:
            dict_only = sub_dict[0]
            dic_tsv_results, Match_id, database_info, dic_temp_db = match_metab_main(
                dict_only,
                key_sub,
                column_name,
                dic_tsv_results,
                Match_id,
                database_info,
                dic_temp_db,
                partial=True,
            )


        return dic_tsv_results, Match_id, database_info, dic_temp_db

    else:
        dic_temp = {}
        if len(sub_dict) > 1:
            # If multiple matches, iterate over all and add matches using met
            for dic in sub_dict:
                dic_tsv_results, Match_id, database_info = match_metab_main(
                    dic,
                    met,
                    column_name,
                    dic_tsv_results,
                    Match_id,
                    database_info,
                    dic_temp,
                    partial=False,
                )

        elif len(sub_dict) == 1:
            dict_only = sub_dict[0]
            dic_tsv_results, Match_id, database_info = match_metab_main(
                dict_only,
                met,
                column_name,
                dic_tsv_results,
                Match_id,
                database_info,
                dic_temp,
                partial=False,
            )

        return dic_tsv_results, Match_id, database_info


def match_met_sbml(
    met,
    meta_data_sbml,
    dictionary_db,
    dic_tsv_results,
    Match_id,
    dic_couple_sbml,
    doublons,
    choice,
    column_name,
    database_info,
):
    """
    Matches a metabolite name from SBML metadata to identify if it
    exists in the metabolic network. If found, the function adds
    relevant information to result dictionaries and checks database
    matching.

    Args:
        met (str): Metabolite name to search for.
        meta_data_sbml (dict): Metadata extracted from SBML models.
        dic_tsv_results (list): Accumulator for harmonized result
          dictionaries.
        Match_id (dict): Dictionary tracking matched metabolites and
          their IDs.
        dic_couple_sbml (dict): Maps SBML filenames to lists of
          metabolite IDs.
        doublons (list): List of metabolites matched in the SBML to
          detect duplicates.
        choice (str): Either 'community' or 'none', defines the mode
          of matching.
        column_name (str): The name of the column used for matching
          (e.g. 'COMMON-NAME').
        database_info (list): Accumulator for matched values for
          logging or traceability.

    Returns:
        tuple: Updated dic_tsv_results, Match_id, doublons, and
        database_info.
    """
    dic_temp = {}
    temp_list = []

    # Find sub-dictionary that contains the metabolite
    sub_dict = utils.find_all_sub_dicts_by_nested_value(meta_data_sbml, met)
    

    if not sub_dict:
        return dic_tsv_results, Match_id, doublons, database_info
    

    # Extract all values from the ‘ID’ key
    id_unique_sbml = [ id_val for sub_sub_dict in sub_dict if "ID" in sub_sub_dict for id_val in sub_sub_dict["ID"]]
    
    # Merge with ‘_AND_’
    id_unique_sbml_list = " _AND_ ".join(id_unique_sbml)

    # COMMUNITY MODE — associate match with SBML filename
    if choice == "community":
        for id_unique in id_unique_sbml:
            sbml_name = utils.find_key_by_list_value(dic_couple_sbml, id_unique)
            temp_list.append(sbml_name)
            logger.info(
                f'--"{met}" is present directly in "{sbml_name}" metabolic network '
                f'with the ID "{id_unique}" via "{column_name}"'
            )
        dic_temp["Match in metabolic networks"] = list(set(temp_list))
    else:
        # CLASSIC MODE — only store ID
        if len(id_unique_sbml) > 1:
            dic_temp["Partial match"] = id_unique_sbml_list
            logger.info(
            f'--""{met}"" has a partial match. We have match for '
            f'more than one id in metabolic network: "{id_unique_sbml}"'
        )
        dic_temp["Match in metabolic networks"] = id_unique_sbml
        logger.info(
            f'--"{met}" is present directly in metabolic network with the ID '
            f'"{id_unique_sbml}" via "{column_name}"'
        )

    # Add to dic_temp regardless the MODE
    if choice == "community" :
        dic_temp["Metabolites"] = f"{met}"
    else:
        dic_temp["Metabolites"] = f"{met} _AND_ {id_unique_sbml_list}"
    dic_temp[f"Match via {column_name}"] = "YES"
    dic_temp["Match in database"] = ""
    Match_id[met] = "NO UNIQUE-ID"
    doublons.append(met)

    # # If we find a match by formula instead of proper ID — log it as partial
    for subform in sub_dict:
        if utils.check_formula_in_dict(subform, met):
            dic_temp["Partial match"] = met
            logger.info(
                f'--""{met}"" has a partial match. We have a formula '
                f'as identifier for this metabolite: "{met}"'
            )

    # Add to result if new entry was created
    if dic_temp:
        dic_tsv_results.append(dic_temp)
    # Try to complete with info from the database
    dic_tsv_results, Match_id, database_info = identification_sbml_match_into_db(
        dictionary_db,
        met,
        column_name,
        dic_tsv_results,
        Match_id,
        database_info,
        key_sub=None,
        partial=False,
    )
    return dic_tsv_results, Match_id, doublons, database_info


def match_db_sbml(
    met,
    meta_data_sbml,
    dic_tsv_results,
    dic_couple_sbml,
    db_list,
    set_list,
    temp_list,
    choice,
):
    """
    Matches a database metabolite identifier to an SBML network and
    updates result records accordingly.

    Args:
        met (str): The metabolite ID or name used to search in SBML metadata.
        meta_data_sbml (dict): Metadata dictionary extracted from SBML models.
        dic_tsv_results (list of dict): List of dictionaries storing matching results.
        dic_couple_sbml (dict): Maps SBML file names to metabolite IDs.
        db_list (list): List of metabolites found in the database to check against.
        set_list (list): Keys used to identify the relevant dictionary
          in dic_tsv_results.
        temp_list (set): Set used to collect matched SBML file names
          (for community mode).
        choice (str): Mode of operation, either 'community' or 'none'.

    Returns:
        dic_tsv_results (list of dict): List of dictionaries storing matching results.
    """
    sub_dict = utils.find_all_sub_dicts_by_nested_value(meta_data_sbml, met) 


    if not sub_dict:
        return dic_tsv_results
 

    # Extract all values from the ‘ID’ key
    id_unique_sbml = [ id_val for sub_sub_dict in sub_dict if "ID" in sub_sub_dict for id_val in sub_sub_dict["ID"]]

    sub_results_dic = utils.find_matching_dict_all_key(dic_tsv_results, set_list)
    sbml_names = utils.find_keys_with_value_in_dict(dic_couple_sbml, met)

    temp_list.add(tuple(sbml_names))  # Add SBML names as a tuple for hashability


    if not sub_results_dic or "Metabolites" not in sub_results_dic:
        return dic_tsv_results
    # Match each metabolite listed in the result dictionary
    metabolites = [m.strip() for m in sub_results_dic["Metabolites"].split(" _AND_ ")]

    for id_unique in id_unique_sbml:
        if id_unique not in metabolites:
            # COMMUNITY MODE — associate match with SBML filename
            if choice == "community":
                if sbml_names:
                    logger.info(
                        f'--"{db_list[-1]}" is present in {sbml_names} '
                        f"metabolic network with the corresponding ID "
                        f'"{id_unique}" via the match ID "{met}"'
                    )
                    if sub_results_dic["Metabolites"]:
                        sub_results_dic["Metabolites"]= (
                                f'{sub_results_dic["Metabolites"]} _AND_ {id_unique}'
                            )
                    else:
                        sub_results_dic["Metabolites"]= id_unique
                    # Flatten and deduplicate SBML name list
                    flat_list = sorted(
                        set(
                            item
                            for group in temp_list
                            for item in (
                                group if isinstance(group, (list, tuple)) else [group]
                            )
                        )
                    )
                    if sub_results_dic.get("Match in metabolic networks"):
                        sub_results_dic["Match in metabolic networks"] = list(dict.fromkeys(sub_results_dic["Match in metabolic networks"] + flat_list))
                    else:
                        sub_results_dic["Match in metabolic networks"] =  flat_list 
            else:
                # CLASSIC MODE — only store ID
                if sub_results_dic.get("Match in metabolic networks"):
                    sub_results_dic["Match in metabolic networks"].append(id_unique)
                    logger.info(
                    f'--""{met}"" has a partial match. We have match for '
                    f'more than one id in metabolic network: "{sub_results_dic["Match in metabolic networks"]}"'
                    )
                else:
                    sub_results_dic["Match in metabolic networks"] = [id_unique]
                if sub_results_dic["Metabolites"]:
                    sub_results_dic["Metabolites"]= (
                            f'{sub_results_dic["Metabolites"]} _AND_ {id_unique}'
                        )
                else: 
                    sub_results_dic["Metabolites"]=id_unique
                logger.info(
                    f'--"{db_list[-1]}" is present directly in '
                    f"metabolic network with the corresponding ID "
                    f'"{id_unique}" via the match ID "{met}"'
                )
            # Check for formula-based partial match
            for subform in sub_dict:
                if utils.check_formula_in_dict(subform, met):
                    sub_results_dic["Partial match"] = met
                    logger.info(
                        f'--""{met}"" has a partial match. We have a '
                    f'formula as identifier for this metabolite: "{met}"'
                )
    
    return dic_tsv_results




def setup_match_db_sbml(
    meta_data_sbml,
    dic_tsv_results,
    dictionary_db,
    dic_couple_sbml,
    database_info,
    doublons,
    choice,
):
    """
    Performs matching between database metabolites and SBML metabolic networks.

    Args:
        meta_data_sbml (dict): Metadata of the SBML files, containing metabolite IDs.
        dic_tsv_results (list of dict): Current list of match results for metabolites.
        dic_couple_sbml (dict): Dictionary with SBML file names as
          keys and lists of metabolites as values.
        database_info (list of list): Each sublist contains all known
          IDs or names for a single metabolite.
        duplicates (list): List of metabolite IDs already matched (to
          avoid re-processing).
        choice (str): Either 'community' or 'none', determining the processing mode.

    Returns:
        dic_tsv_results (list of dict): Updated `dic_tsv_results` with
          any new matches found.
    """
    # Loop through each set of identifiers for one metabolite
    for db_list in database_info:
        temp_list = set()  # To store matched SBML networks temporarily
        set_list = db_list[-1]  # Usually the last item is the 'UNIQUE-ID'
        
        # Skip if this metabolite was already matched
        if set_list not in doublons:
            # Loop through all possible names/IDs for this metabolite
            for met_info in db_list:
                dic_tsv_results = match_db_sbml(
                    met_info,
                    meta_data_sbml,
                    dic_tsv_results,
                    dic_couple_sbml,
                    db_list,
                    set_list,
                    temp_list,
                    choice,
                )
    return dic_tsv_results


# ------------------------------------------------------------#
#             Partial Match Metabolites                      #
# ------------------------------------------------------------#


def partial_match_met_sbml(
    met,
    key_sub,
    meta_data_sbml,
    dictionary_db,
    dic_tsv_results,
    Match_id,
    dic_couple_sbml,
    doublons,
    choice,
    column_name,
    database_info,
    dic_temp,
    temp_list,
):
    """
    Handle partial matches between a metabolite and the SBML metabolic networks.

    Searches for partial matches of `met` within `meta_data_sbml`. If
    found, updates the `dic_tsv_results` with partial match
    information, depending on the mode specified by `choice`.

    In 'community' mode, collects all metabolic network names containing the match.
    In standard mode, records a direct match with the unique SBML ID.

    Also updates `Match_id`, `doublons`, and attempts to integrate
    this match into the database matching results.

    Args:
        met (str): Metabolite identifier to search for partial matches.
        key_sub (str): Original metabolite key from the input dataset.
        meta_data_sbml (dict): Dictionary containing SBML metabolite metadata.
        dic_tsv_results (list of dict): List of dictionaries storing match results.
        Match_id (dict): Dictionary mapping metabolite keys to matched IDs.
        dic_couple_sbml (dict): Maps SBML network names to lists of metabolites.
        doublons (list): List tracking metabolites already matched or
          identified as duplicates.
        choice (str): Mode of operation; 'community' or 'none'.
        column_name (str): Column name used to annotate how the match was made.
        database_info (list): List containing database matching metadata.
        dic_temp (dict): Temporary dictionary for current metabolite match info.
        temp_list (list): Temporary list of matched SBML network names
          (used in community mode).

    Returns:
        tuple: Updated (dic_tsv_results, Match_id, doublons,
        database_info, dic_temp, temp_list).
    """
    sub_dict = utils.find_all_sub_dicts_by_nested_value(meta_data_sbml, met)


    if not sub_dict:
        return dic_tsv_results, Match_id, doublons, database_info, dic_temp, temp_list

    # Extract base unique SBML ID
    id_unique_sbml = [ id_val for sub_sub_dict in sub_dict if "ID" in sub_sub_dict for id_val in sub_sub_dict["ID"]]

    if choice == "community":
        for id_unique in id_unique_sbml:
            sbml_name = utils.find_key_by_list_value(dic_couple_sbml, id_unique)
            temp_list.append(sbml_name)
        dic_temp["Match in metabolic networks"] = list(set(temp_list))
    else:
        dic_temp["Match in metabolic networks"] = id_unique_sbml
        dic_temp[f"Match via {column_name}"] = "YES"

    dic_temp["Metabolites"] = key_sub
    Match_id[key_sub] = "NO UNIQUE-ID"

    # Append or extend the partial match string efficiently
    previous_partial = dic_temp.get("Partial match", "")
    dic_temp["Partial match"] = (
        met if not previous_partial else f"{previous_partial} _AND_ {met}"
    )

    logger.info(
        f'      --"{key_sub}" has a partial match. A match was found '
        f'with the ID: "{met}" via: {column_name}'
    )

    # Append dic_temp only if new to results
    if dic_temp not in dic_tsv_results:
        dic_tsv_results.append(dic_temp)
        doublons.append(key_sub)

    # Update matches further in the database (partial mode)
    dic_tsv_results, Match_id, database_info = identification_sbml_match_into_db(
        dictionary_db,
        met,
        column_name,
        dic_tsv_results,
        Match_id,
        database_info,
        key_sub,
        "partial",
    )
    return dic_tsv_results, Match_id, doublons, database_info, dic_temp, temp_list





# ---------------------------------------#
#               MAPPING                 #
# ---------------------------------------#


def mapping_run(
    output_folder,
    dictionary_db,
    maf_dictionnary,
    keys,
    maf_df,
    meta_data_sbml,
    dic_couple_sbml,
    start_time,
    partial_match,
    quiet,
    timestamp,
    choice,
):
    """
    Centralises the mapping step, including data preparation,
    matching/unmatching processes, and result compilation. At the end,
    generates a TSV file with the mapping results.

    Args:
        output_folder (str or Path): Path to the output directory.
        dictionary_db (dict): Dictionary where each key corresponds to
          a metabolite from the MetaCyc/MetaNetX database and the values contain
          information such as IDs from other databases.
        maf_dictionnary (dict): Dictionary containing data from the
          MAF file used for matching against the conversion datatable.
        maf_df (pandas.DataFrame): DataFrame version of the MAF file,
          used for extracting CHEBI/InChIKeys.
        columns_name_maf (list): List of relevant column names from
          the MAF dictionary (after filtering out empty or irrelevant
          columns).
        keys (list): List of all possible column names expected in the
          output or input files, used to standardize/harmonize the final
          output.
        List_merge_sbml_metabolites (list): List of all metabolites
          present in the SBML files.
        meta_data_sbml (dict): Metadata extracted from the SBML files
          (e.g., IDs, names, formulas).
        dic_couple_sbml (dict): Dictionary where each key is an SBML
          file name and each value is the list of metabolites it
          contains.
        start_time (int or str): Timestamp marking the start of
          execution, used for naming output files uniquely.
        timestamp (int): Begin time of the run
        partial_match (str):  Defines if the script is run
          in partial_match option or not.
        choice (str): Either 'community' or 'none'. Defines if the
          script is run in community mode or classic mode.

    Returns:
        None. Writes output to a TSV file in the specified output folder.
    """

    # ---------------------------#
    #  Set output format part   #
    # ---------------------------#

    # Results file
    Match_id = {}
    doublons = []
    database_info = []
    dic_tsv_results = []

    List_maf_metabolites = list(itertools.chain.from_iterable(maf_dictionnary.values()))
    List_maf_metabolites_wt_nan = list(
        dict.fromkeys([x for x in List_maf_metabolites if x != "nan"])
    )

    # ---------------#
    #  Match part   #
    # ---------------#

    logger.info("\n\n----------------------------------------------")
    logger.info("---------------MATCH STEP 1-------------------")
    logger.info("----------------------------------------------\n")

    logger.info(
        "<1> Direct matching test between metabolites derived from "
        "metabolomic data on all metadata in the metabolic network"
    )
    logger.info(
        "<2> Matching test between metabolites derived from "
        "metabolomic data on all metadata in the database conversion"
    )

    for column_name, value in maf_dictionnary.items():
        list_maf_metabolites_wt_nan = [x for x in value if x != "nan"]
        for met in list_maf_metabolites_wt_nan:
            logger.info(f'\n++ Match step for "{met}":')
            dic_tsv_results, Match_id, doublons, database_info = match_met_sbml(
                met,
                meta_data_sbml,
                dictionary_db,
                dic_tsv_results,
                Match_id,
                dic_couple_sbml,
                doublons,
                choice,
                column_name,
                database_info,
            )
            dic_tsv_results, Match_id, database_info = match_metabo(
                dictionary_db,
                met,
                column_name,
                dic_tsv_results,
                Match_id,
                database_info,
                key_sub=None,
                dic_temp_db=None,
                partial=False,
            )

    logger.info("\n\n----------------------------------------------")
    logger.info("---------------MATCH STEP 2-------------------")
    logger.info("----------------------------------------------\n")

    logger.info(
        "<3> Matching test on metabolites that matched only on the "
        "database conversion data against all metadata from the "
        "metabolic network\n"
    )

    dic_tsv_results = setup_match_db_sbml(
        meta_data_sbml,
        dic_tsv_results,
        dictionary_db,
        dic_couple_sbml,
        database_info,
        doublons,
        choice,
    )

    # Groups similar dictionaries based on shared values in key fields, then merges
    # them.
    dic_tsv_results = utils.smart_merge(dic_tsv_results)
    if choice != "community":
        for dic in dic_tsv_results:
            if dic.get("Match in metabolic networks"):
                if len(dic["Match in metabolic networks"]) > 1 :
                    if dic.get("Partial match"):
                        id_unique_sbml_list = " _AND_ ".join(dic["Match in metabolic networks"])
                        merge = " _AND_ ".join(
                            dict.fromkeys([dic["Partial match"], id_unique_sbml_list])
                            )
                        dic["Partial match"]=  merge
                        logger.info(
                            f'--Partial match. We have match for '
                            f'more than one id in metabolic network: "{merge}"'
                            )
                    else:
                        id_unique_sbml_list = " _AND_ ".join(dic["Match in metabolic networks"])
                        dic["Partial match"]=  id_unique_sbml_list
                    logger.info(
                    f'--Partial match. We have match for '
                    f'more than one id in metabolic network: "{dic["Match in metabolic networks"]}"'
        )

    if partial_match:
        logger.info("\n\n------------------------------------------------")
        logger.info("---------------PARTIAL MATCH -------------------")
        logger.info("----------------------------------------------\n")
        logger.info(
            "<4> Attempting to find new IDs for unmatched metabolites "
            "using multiple strategies: removing enantiomer "
            "specificity, using the core part of InChIKeys, or "
            "exploring parent/child relationships in CHEBI.\n"
        )

        dic_unmatch_to_reload = {}

        # Filter metabolites absent in both database and metabolic network
        unmatch_metabolites_try = list(
            set(List_maf_metabolites_wt_nan) - set(Match_id.keys())
        )
        unmatch_metabolites = utils.process_unmatches(
            maf_df, set(Match_id.keys()), unmatch_metabolites_try
        )

        list_unmatch_to_reload = {}
        if any(col.lower() == "chebi" for col in maf_df.columns):
            logger.info("\n-----> CHEBI <-----")
            # Extract the chebi from the maf datatable merge
            dic_chebi_match = utils.find_targets_with_chebi(maf_df, unmatch_metabolites)

            list_unmatch_to_reload = asyncio.run(utils.chebi_ontology_main_process(dic_chebi_match))
            dic_unmatch_to_reload["CHEBI"] = list_unmatch_to_reload


        logger.info("\n-----> inchikey <-----")
        dic_Inchey_match = {
            i: [i.split("-")[0]]
            for i in unmatch_metabolites
            if "inchikey" in str(i).lower()
        }
        for key, value in dic_Inchey_match.items():
            logger.info(f"inchikey: {key} -> Short inchikey: {value}")
        dic_unmatch_to_reload["inchikey"] = dic_Inchey_match

        # ------------------------------------------------------------------------- #
        # Remove the L - D - R -S  enantiomerS if the opt is true and
        # update the datatable conversion temporary for the mapping.
        # ------------------------------------------------------------------------- #
        logger.info("\n-----> Enantiomers <-----")
        dic_enantiomers_match = list(
            set(unmatch_metabolites)
            - set(list_unmatch_to_reload.keys())
            - set(dic_Inchey_match.keys())
        )
        dic_unmatch_to_reload["Enantiomers"] = dic_enantiomers_match
        logger.info(
            "-----> Remove enantiomers in database and in metadata of sbml <-----"
        )

        dictionary_db_enantiomers = remove_enantiomer_and_Inchey_db(dictionary_db)
        dictionary_metadata_enantiomers = remove_enantiomer_and_Inchey_metadata(
            meta_data_sbml
        )

        logger.info("\n\n----------------------------------------------------")
        logger.info("---------------Retry MATCH STEP 1-------------------")
        logger.info("----------------------------------------------------")
        database_info = []

        for column_name, value in dic_unmatch_to_reload.items():
            if column_name == "CHEBI":
                logger.info("\n-----> Test: CHEBI <-----")
                for key_sub, val_sub in value.items():
                    dic_temp_sbml = {}
                    dic_temp_db = {}
                    temp_list = []
                    dic_temp_sbml["Partial match"] = ""
                    dic_temp_sbml["Match in database"] = ""
                    logger.info(f'\n++ Match step for "{key_sub}":')
                    for met in val_sub:
                        logger.info(f'\n--Try with {met}":')
                        (
                            dic_tsv_results,
                            Match_id,
                            doublons,
                            database_info,
                            dic_temp_sbml,
                            temp_list,
                        ) = partial_match_met_sbml(
                            met,
                            key_sub,
                            meta_data_sbml,
                            dictionary_db,
                            dic_tsv_results,
                            Match_id,
                            dic_couple_sbml,
                            doublons,
                            choice,
                            column_name,
                            database_info,
                            dic_temp_sbml,
                            temp_list,
                        )
                        dic_tsv_results, Match_id, database_info, dic_temp_db = (
                            match_metabo(
                                dictionary_db,
                                met,
                                column_name,
                                dic_tsv_results,
                                Match_id,
                                database_info,
                                key_sub,
                                dic_temp_db,
                                "partial",
                            )
                        )
            elif column_name == "inchikey":
                logger.info("\n-----> Test: Inchey <-----")
                for key_sub, val_sub in value.items():
                    dic_temp_sbml = {}
                    dic_temp_db = {}
                    temp_list = []
                    met = val_sub[0]
                    logger.info(f'\n++ Match step for "{met}":')
                    logger.info(f'\n--Try with {key_sub}":')
                    (
                        dic_tsv_results,
                        Match_id,
                        doublons,
                        database_info,
                        dic_temp_sbml,
                        temp_list,
                    ) = partial_match_met_sbml(
                        met,
                        key_sub,
                        dictionary_metadata_enantiomers,
                        dictionary_db,
                        dic_tsv_results,
                        Match_id,
                        dic_couple_sbml,
                        doublons,
                        choice,
                        column_name,
                        database_info,
                        dic_temp_sbml,
                        temp_list,
                    )
                    dic_tsv_results, Match_id, database_info, dic_temp_db = (
                        match_metabo(
                            dictionary_db_enantiomers,
                            met,
                            column_name,
                            dic_tsv_results,
                            Match_id,
                            database_info,
                            key_sub,
                            dic_temp_db,
                            "partial",
                        )
                    )
            else:
                logger.info("\n-----> Test: Enantiomers <-----")
                for met in value:
                    dic_temp_sbml = {}
                    dic_temp_db = {}
                    temp_list = []
                    logger.info(f'\n++ Match step for "{met}":')
                    (
                        dic_tsv_results,
                        Match_id,
                        doublons,
                        database_info,
                        dic_temp_sbml,
                        temp_list,
                    ) = partial_match_met_sbml(
                        met,
                        met,
                        dictionary_metadata_enantiomers,
                        dictionary_db,
                        dic_tsv_results,
                        Match_id,
                        dic_couple_sbml,
                        doublons,
                        choice,
                        column_name,
                        database_info,
                        dic_temp_sbml,
                        temp_list,
                    )
                    dic_tsv_results, Match_id, database_info, dic_temp_db = (
                        match_metabo(
                            dictionary_db_enantiomers,
                            met,
                            column_name,
                            dic_tsv_results,
                            Match_id,
                            database_info,
                            met,
                            dic_temp_db,
                            "partial",
                        )
                    )

        logger.info("\n\n----------------------------------------------------")
        logger.info("---------------Retry MATCH STEP 2-------------------")
        logger.info("----------------------------------------------------")

        dic_tsv_results = setup_match_db_sbml(
            meta_data_sbml,
            dic_tsv_results,
            dictionary_db,
            dic_couple_sbml,
            database_info,
            doublons,
            choice,
        )
        dic_tsv_results = setup_match_db_sbml(
            dictionary_metadata_enantiomers,
            dic_tsv_results,
            dictionary_db,
            dic_couple_sbml,
            database_info,
            doublons,
            choice,
        )

        # Groups similar dictionaries based on shared values in key
        # fields, then merges them.
        dic_tsv_results = utils.smart_merge(dic_tsv_results)
        

    # #    #----------------------------#
    # #    #  Umatch set up the output  #
    # #    #----------------------------#
    unmatch_metabolites_try_last = list(
        set(List_maf_metabolites_wt_nan) - set(Match_id.keys())
    )
    unmatch_metabolites_total = utils.process_unmatches(
        maf_df, set(Match_id.keys()), unmatch_metabolites_try_last
    )
    unmatch_metabolites_merged = utils.extract_metabolite_combinations(
        unmatch_metabolites_total, maf_df
    )

    keys_starter = [
        "Metabolites",
        "Match in database",
        "Match in metabolic networks",
        "Partial match",
    ]
    dic_tsv_results, keys_reorder = setup_harmonisation_output(
        dic_tsv_results, keys_starter, unmatch_metabolites_merged, keys
    )

    #      #-----------------------------#
    # #    #   Results User Interface    #
    # #    #-----------------------------#

    stats = utils.analyze_column_matches(
        dict_list=dic_tsv_results,
        col_db="Match in database",
        col_net="Match in metabolic networks",
        col_partial="Partial match",
    )
    # Terminal results
    logger.info("\n-------------------- SUMMARY REPORT --------------------\n")

    logger.info("\nRecap of Matches:")
    logger.info("  + Matched metabolites: %s", stats["network_total_filled"])
    logger.info("  + Unmatched metabolites: %s", stats["network_empty_total"])
    logger.info(
        "  + Partial matches: %s",
        stats["partial_and_metabolic_filled"] + stats["db_and_partial_only"],
    )

    logger.info("\n Match Details:")
    logger.info(f"  --> Full match (database + SBML): {stats['both_filled']}")
    logger.info(
        f"  --> Partial match + metabolic info: {stats['partial_and_metabolic_filled']}"
    )
    logger.info(f"  --> Match only in SBML: {stats['network_only']}")

    logger.info("\n Unmatch Details:")
    logger.info(f"  --> Full unmatch (no match in DB or SBML): {stats['both_empty']}")
    logger.info(f"  --> Match in DB but not in SBML: {stats['db_only']}")
    logger.info(f"  --> Partial match in DB only: {stats['db_and_partial_only']}")

    logger.info("\n--------------------------------------------------------\n")

    # Write output file in funtion of the "choice"
    if partial_match:
        if choice == "community":
            utils.write_tsv(
                dic_tsv_results,
                output_folder,
                f"{choice}_mapping_results_partial_match_{start_time}.tsv",
                keys_reorder,
                quiet,
            )  # Write results
        else:
            utils.write_tsv(
                dic_tsv_results,
                output_folder,
                f"mapping_results_partial_match_{start_time}.tsv",
                keys_reorder,
                quiet,
            )  # Write results
    else:
        if choice == "community":
            utils.write_tsv(
                dic_tsv_results,
                output_folder,
                f"{choice}_mapping_results_{start_time}.tsv",
                keys_reorder,
                quiet,
            )  # Write results
        else:
            utils.write_tsv(
                dic_tsv_results,
                output_folder,
                f"mapping_results_{start_time}.tsv",
                keys_reorder,
                quiet,
            )  # Write results
    if quiet:
        print(
            "\n--- Total runtime %.2f seconds ---\n ---> MAPPING COMPLETED'"
            % (time.time() - timestamp)
        )



#!/bin/python
## MISTIC Project INRIA/INRAE
## Author Muller Coralie
## Date: 2025/08/19
## Update: 2025/10/31

"""
Description:
Test build datatable function for Metacyc and MetNetX mode
"""
from os import path
import pytest
from pathlib import Path
import csv
import os
import tempfile
import json
import pandas as pd
from argparse import Namespace
from metanetmap import build_database
from unittest.mock import patch



#------------------------------------#
#        DIRECTORIES AND FILES       #
#------------------------------------#
TEST_TOYS_DIR = Path(__file__).parent.parent

DATATABLE_COMPLEMENT_METACYC = path.join(TEST_TOYS_DIR,'src/metanetmap/build_datatable_conversion/datatable_complementary_metacyc.tsv')
DATATABLE_COMPLEMENT_METANETX = path.join(TEST_TOYS_DIR,'src/metanetmap/build_datatable_conversion/datatable_complementary_metanetx.tsv')
TEST_EXPECTED_DIR = Path(__file__).parent

# ----------------------------------------------------------

#utils
def read_tsv(fp):
    with open(fp, encoding='utf-8') as f:
        return sorted(list(csv.DictReader(f, delimiter='\t')), key=lambda x: str(x))
    

# ----------------------------------------------------------

#-----------------------#
#         TESTS         #
#-----------------------#

@pytest.mark.skipif(
    not os.path.exists("tests/data/compounds_29.dat") or
    not os.path.exists("tests/data/conversion_datatable.tsv"),
    reason="Required test data files are missing"
)
# Main --> Compare the datatable generated to the one expected 
#### Results table
def test_build_datatable_metacyc(tmp_path):
    # 1. Settings
    output_folder = tmp_path
    output_folder.mkdir(parents=True, exist_ok=True) 
    output_path=Path(output_folder,"conversion_datatable.tsv")

    # 2. Load input data
    Metacyc = "tests/data/compounds_29.dat"
    CONVERSION_DATATABLE = "tests/data/conversion_datatable.tsv"

    # 3. Run the build step
    args = Namespace(
        metacyc_file=Metacyc,
        complement_file=DATATABLE_COMPLEMENT_METACYC,
        output=output_path,
        db='metacyc',
        quiet=True,
                )
    build_database.load_args(args)
   
    
    # 4. Check the generated file
    assert output_path.exists(), "Output file not generated"

    # 5.  Compare contents
    actual = read_tsv(output_path)
    expected = read_tsv(CONVERSION_DATATABLE)
    assert actual == expected, "The generated table does not match the expected one"






@pytest.mark.skipif(
    not os.path.exists("tests/data/chem_xref.tsv") or
    not os.path.exists("tests/data/chem_prop.tsv") or
    not os.path.exists("tests/data/metanetx_conversion_datatable.tsv"),
    reason="Required test data files are missing"
)
# Main --> Compare the datatable generated to the one expected 
#### Results table
def test_build_datatable_metanetx(tmp_path):
    # 1. Settings
    output_folder = tmp_path
    output_folder.mkdir(parents=True, exist_ok=True) 
    output_path=Path(output_folder,"conversion_datatable.tsv")

    # 2. Load input data
    chem_prop_file="tests/data/chem_prop.tsv"
    chem_ref_file="tests/data/chem_xref.tsv"
    CONVERSION_DATATABLE_METANETX = "tests/data/metanetx_conversion_datatable.tsv"

    # 3. Run the build step
    args = Namespace(
        chem_prop_file=chem_prop_file,
        chem_ref_file=chem_ref_file,
        complement_file=DATATABLE_COMPLEMENT_METANETX,
        output=output_path,
        db='metanetx',
        quiet=True,
    )
    build_database.load_args(args)

    
    # 3. Check the generated file
    assert output_path.exists(), "Output file not generated"

    # 4.  Compare contents
    actual = read_tsv(output_path)
    expected = read_tsv(CONVERSION_DATATABLE_METANETX)
    assert actual == expected, "The generated table does not match the expected one"




def test_split_line_adds_extracted_value_to_dict():
    # Simulate a binary line from a MetaCyc/MetaNetX file
    line = b'(COMMON-NAME "Glucose")'
    column_name = "COMMON-NAME"
    dictionary_temp = {}

    # Call the function
    build_database.split_line(line, column_name, dictionary_temp)

    # Check if the dictionary was updated correctly
    assert column_name in dictionary_temp
    assert dictionary_temp[column_name] == "Glucose"


def test_replace_header_line_updates_dictionary():
    # Example binary line from a MetaCyc/MetaNetX file header
    line = b'UNIQUE-ID - META00001\n'
    column_name = "UNIQUE-ID"
    dictionary_temp = {}

    # Call the function
    build_database.replace_header_line(line, column_name, dictionary_temp)

    # Assertions
    assert column_name in dictionary_temp
    assert dictionary_temp[column_name] == "META00001"



def test_replace_line_cleans_and_simplifies_text():
    # Example raw MetaCyc/MetaNetX line in bytes
    line = b'COMMON-NAME - <i>D-glucose</i> &rarr; energy\n'

    # Expected cleaned result
    expected = "D-glucose -> energy"

    # Call the function
    result = build_database.replace_line(line)

    # Assertion
    assert result == expected



def test_build_main_dictionary_creates_correct_structure():
    # Create example MetaCyc content as bytes (mock file content)
    content = b"""
UNIQUE-ID - CPD-17659
COMMON-NAME - D-allo-isoleucine
SYNONYMS - isoleucine-allo
ABBREV-NAME - 
MOLECULAR-WEIGHT - 131.17
MONOISOTOPIC-MW - 131.095
DBLINKS - (CHEBI "CHEBI:85306")
DBLINKS - (SEED "C00022")
DBLINKS - (BIGG "bigg_id")
DBLINKS - (HMDB "HMDB0005760"
DBLINKS - (METANETX "MNXM17053")
DBLINKS - (LIGAND-CPD "C21092")
DBLINKS - (REFMET "Lyxose")
DBLINKS - (PUBCHEM "897")
DBLINKS - (CAS "30237-26-4")
INCHI-KEY - InChIKey=AGPKZVBTJJNPAG-CRCLSJGQSA-N
SMILES - CC[C@H](C)[C@@H]([NH3+])C([O-])=O
//
UNIQUE-ID - CPD-12345
COMMON-NAME - glucose
DBLINKS - (CHEBI "CHEBI:12345")

"""

    # Write the content to a temporary file
    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        tmpfile.write(content)
        tmpfile_path = tmpfile.name

    # Call the function
    dictionary_db, keys = build_database.build_main_dictionary(tmpfile_path)

    # Clean up the temp file
    os.remove(tmpfile_path)

    # Check the output structure
    assert isinstance(dictionary_db, list)
    assert len(dictionary_db) == 2  # Two metabolites in example

    first_entry = dictionary_db[0]
    assert first_entry['UNIQUE-ID'] == "CPD-17659"
    assert first_entry['CHEBI'] == "CHEBI:85306"  # Assuming split_line strips 'CHEBI:'
    assert 'D-allo-isoleucine' in first_entry['COMMON-NAME']
    assert 'isoleucine-allo' in first_entry['SYNONYMS']
    assert first_entry['MOLECULAR-WEIGHT'] == "131.17"
    assert first_entry['INCHI-KEY'].startswith("InChIKey=")
    assert keys[0] == "UNIQUE-ID"

    second_entry = dictionary_db[1]
    assert second_entry['UNIQUE-ID'] == "CPD-12345"
    assert second_entry['CHEBI'] == "CHEBI:12345"
    assert second_entry['COMMON-NAME'] == "glucose"



def test_manage_synonyms():
    # Input with list of synonyms
    input_data = [
        {'UNIQUE-ID': 'CPD-1', 'SYNONYMS': ["syn1", "syn2"]},
        {'UNIQUE-ID': 'CPD-2', 'SYNONYMS': ""},
        {'UNIQUE-ID': 'CPD-3', 'SYNONYMS': ['single_synonym']},
        {'UNIQUE-ID': 'CPD-4'},  # No SYNONYMS key
    ]
    
    output = build_database.manage_synonyms(input_data)
    
    # Check first entry: list converted to JSON string
    assert isinstance(output[0]['SYNONYMS'], str)
    assert json.loads(output[0]['SYNONYMS']) == ['syn1', 'syn2']

    # Check second entry: None unchanged
    assert output[1]['SYNONYMS'] == ""

    # Check first entry: list converted to JSON string
    assert json.loads(output[2]['SYNONYMS']) == ["single_synonym"]

    # Check fourth entry: no SYNONYMS key remains absent
    assert 'SYNONYMS' not in output[3]


@pytest.mark.parametrize(
    "dictionary_to_add, column_name_add, dictionary_db, expected_output",
    [
        # Test Case 1: Normal case with partial match
        (
            {'CPD-001': 'glc', 'CPD-002': 'pyr'},
            'BIGG',
            [
                {'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose'},
                {'UNIQUE-ID': 'CPD-002', 'COMMON-NAME': 'Pyruvate'},
                {'UNIQUE-ID': 'CPD-003', 'COMMON-NAME': 'Lactate'}
            ],
            [
                {'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose', 'BIGG': 'glc'},
                {'UNIQUE-ID': 'CPD-002', 'COMMON-NAME': 'Pyruvate', 'BIGG': 'pyr'},
                {'UNIQUE-ID': 'CPD-003', 'COMMON-NAME': 'Lactate'}
            ]
        ),
        # Test Case 2: No matches
        (
            {'CPD-999': 'x'},
            'SEED',
            [{'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose'}],
            [{'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose'},
             {'UNIQUE-ID': 'CPD-999', 'SEED': 'x'}]
        ),
        #Test Case 3: Overwrite existing value
        (
            {'CPD-001': 'new_value'},
            'SEED',
            [{'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose', 'SEED': 'old_value'}],
            [{'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose', 'SEED': 'new_value'}]
        )
    ]
)
def test_add_each_complement(dictionary_to_add, column_name_add, dictionary_db, expected_output):
    result = build_database.add_each_complement(dictionary_to_add, column_name_add, dictionary_db)
    assert result == expected_output




@pytest.mark.parametrize(
    "dictionary_db, complementary_dicts, expected_output",
    [
        # Test Case 1: Add multiple complementary columns
        (
            [
                {'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose'},
                {'UNIQUE-ID': 'CPD-002', 'COMMON-NAME': 'Pyruvate'}
            ],
            {
                'BIGG': {'CPD-001': 'glc', 'CPD-002': 'pyr'},
                'SEED': {'CPD-001': 'cpd00027', 'CPD-002': 'cpd00020'},
                'FORMULA':{'CPD-001': 'C8H6R1N10'}
            },
            [
                {'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose', 'BIGG': 'glc', 'SEED': 'cpd00027','FORMULA': "C8H6R1N10" },
                {'UNIQUE-ID': 'CPD-002', 'COMMON-NAME': 'Pyruvate', 'BIGG': 'pyr', 'SEED': 'cpd00020'}
            ]
        ),

        # Test Case 2: Some IDs in complementary_dicts not in dictionary_db
        (
            [
                {'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose'}
            ],
            {
                'BIGG': {'CPD-001': 'glc', 'CPD-999': 'x'},  # CPD-999 ignored
                'SEED': {'CPD-001': 'cpd00027'}
            },
            [
                {'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose', 'BIGG': 'glc', 'SEED': 'cpd00027'},
                {'UNIQUE-ID': 'CPD-999', 'BIGG': 'x'}
            ]
        ),

        # Test Case 3: Empty complementary_dicts (no updates)
        (
            [{'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose'}],
            {},
            [{'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose'}]
        ),

        # Test Case 4: Overwrite existing field
        (
            [{'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose', 'SEED': 'old_value'}],
            {'SEED': {'CPD-001': 'new_value'}},
            [{'UNIQUE-ID': 'CPD-001', 'COMMON-NAME': 'Glucose', 'SEED': 'new_value'}]
        )
    ]
)
def test_add_complement_from_complementary(dictionary_db, complementary_dicts, expected_output):
    result = build_database.add_complement_from_complementary(dictionary_db, complementary_dicts)
    assert result == expected_output


# Test load complement:
#Helper to write a TSV file
def create_temp_tsv(headers, rows):
    tmp = tempfile.NamedTemporaryFile(mode='w+', delete=False, newline='')
    writer = csv.writer(tmp, delimiter='\t')
    writer.writerow(headers)
    writer.writerows(rows)
    tmp.close()
    return tmp.name


#Parametrized test for successful and partial data loads
@pytest.mark.parametrize("headers, rows, expected", [
    (
        ['UNIQUE-ID', 'BIGG', 'SEED'],
        [['CPD-123', 'glc', 'cpd00027'], ['CPD-456', 'pyr', 'cpd00020']],
        {
            'BIGG': {'CPD-123': 'glc', 'CPD-456': 'pyr'},
            'SEED': {'CPD-123': 'cpd00027', 'CPD-456': 'cpd00020'}
        }
    ),
    (
        ['UNIQUE-ID', 'BIGG', 'SEED'],
        [['CPD-123', '', 'cpd00027'], ['CPD-456', 'pyr', '']],
        {
            'BIGG': {'CPD-456': 'pyr'},
            'SEED': {'CPD-123': 'cpd00027'}
        }
    )
])
def test_load_complementary_datatable_success(headers, rows, expected):
    path = create_temp_tsv(headers, rows)
    try:
        result = build_database.load_complementary_datatable(path)
        assert result == expected
    finally:
        os.remove(path)


#Grouped error tests
def test_load_complementary_datatable_errors(monkeypatch):
    # File not found
    with pytest.raises(FileNotFoundError):
        build_database.load_complementary_datatable("nonexistent.tsv")

    # Missing UNIQUE-ID
    path2 = create_temp_tsv(['BIGG', 'SEED'], [['glc', 'cpd00027']])
    try:
        with pytest.raises(OSError, match="Missing 'UNIQUE-ID'"):
            build_database.load_complementary_datatable(path2)
    finally:
        os.remove(path2)


############################################ 
#   Build_data test specific to MetaNetX   #
############################################ 


#Test download
def test_download_metanetx_file_success():
    """Test a successful download using a mock."""
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = "test_file.tsv"
        url = "http://example.com/test_file.tsv"

        # Mock urllib.request.urlretrieve to just create an empty file
        def fake_urlretrieve(u, fpath):
            with open(fpath, "w") as f:
                f.write("mock content")
            return (fpath, None)

        with patch("urllib.request.urlretrieve", side_effect=fake_urlretrieve):
            result = build_database.download_metanetx_file(filename, url, tmpdir)

        assert result is not None
        assert os.path.exists(result)
        with open(result) as f:
            content = f.read()
        assert content == "mock content"

def test_download_metanetx_file_fail():
    """Test download failure returns None."""
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = "test_file.tsv"
        url = "http://example.com/test_file.tsv"

        # Mock urllib.request.urlretrieve to raise an exception
        with patch("urllib.request.urlretrieve", side_effect=Exception("Failed")):
            result = build_database.download_metanetx_file(filename, url, tmpdir)

        assert result is None



# Check reads files
def test_read_chem_prop_basic():
    """Test reading a small chem_prop.tsv content."""
    content = """MNXM1\tGlucose\tRef1\tC6H12O6\t0\t180.16\tInChI1\tKey1\tSMILES1
MNXM2\tATP\tRef2\tC10H16N5O13P3\t-4\t507.18\tInChI2\tKey2\tSMILES2
"""
    with tempfile.NamedTemporaryFile(mode="w+", delete=True) as tmp:
        tmp.write(content)
        tmp.flush()
        df = build_database.read_chem_prop(tmp.name)
    
    assert df.shape[0] == 2
    assert list(df.columns) == ["MNX_ID","COMMON_NAME","REFERENCE","FORMULA","CHARGE","MOLECULAR_WEIGHT","INCHI","INCHI_KEY","SMILES"]
    assert df.iloc[0]["COMMON_NAME"] == "Glucose"


def test_read_chem_xref_basic():
    """Test reading a small chem_xref.tsv content."""
    content = """chebi:1234\tMNXM1\tGlucose
seed:cpd00001\tMNXM2\tATP
"""
    with tempfile.NamedTemporaryFile(mode="w+", delete=True) as tmp:
        tmp.write(content)
        tmp.flush()
        df = build_database.read_chem_xref(tmp.name)
    
    assert df.shape[0] == 2
    assert list(df.columns) == ["source","MNX_ID","description"]
    assert df.iloc[1]["source"] == "seed:cpd00001"


# extract ids 

def test_extract_ids_basic():
    """Basic test: extract and group database-specific IDs."""
    df = pd.DataFrame({
        "MNX_ID": ["MNXM1", "MNXM1", "MNXM2", "MNXM3"],
        "source": ["chebi:1234", "chebi:5678", "chebi:9999", "seed:0001"]
    })
    
    result = build_database.extract_ids(df, "chebi")
    
    expected = pd.DataFrame({
        "MNX_ID": ["MNXM1", "MNXM2"],
        "CHEBI": ["chebi:1234|chebi:5678", "chebi:9999"]
    })
    
    pd.testing.assert_frame_equal(result.reset_index(drop=True), expected)


def test_extract_ids_duplicates():
    """Duplicate IDs should be deduplicated and sorted."""
    df = pd.DataFrame({
        "MNX_ID": ["MNXM1", "MNXM1", "MNXM1"],
        "source": ["chebi:2", "chebi:1", "chebi:2"]
    })
    
    result = build_database.extract_ids(df, "chebi")
    expected = pd.DataFrame({
        "MNX_ID": ["MNXM1"],
        "CHEBI": ["chebi:1|chebi:2"]
    })
    
    pd.testing.assert_frame_equal(result, expected)

def test_extract_ids_mixed_prefixes():
    """Only rows with the correct prefix should be extracted."""
    df = pd.DataFrame({
        "MNX_ID": ["MNXM1", "MNXM1", "MNXM2"],
        "source": ["chebi:123", "hmdb:456", "chebi:789"]
    })
    
    result = build_database.extract_ids(df, "chebi")
    expected = pd.DataFrame({
        "MNX_ID": ["MNXM1", "MNXM2"],
        "CHEBI": ["chebi:123", "chebi:789"]
    })
    
    pd.testing.assert_frame_equal(result.reset_index(drop=True), expected)




# Explode test 
def test_explode_basic():
    """Basic test: explode a column with '|' separators."""
    df = pd.DataFrame({
        "id": [1, 2],
        "metabolite": ["A|B|C", "D"]
    })
    
    result = build_database.explode_ids(df, "metabolite")
    
    expected = pd.DataFrame({
        "id": [1, 1, 1, 2],
        "metabolite": ["A", "B", "C", "D"]
    })
    
    pd.testing.assert_frame_equal(result.reset_index(drop=True), expected)

def test_explode_ignore_empty():
    """Empty strings or spaces should be ignored."""
    df = pd.DataFrame({
        "id": [1],
        "metabolite": ["A||B| |C|"]
    })
    
    result = build_database.explode_ids(df, "metabolite")
    
    expected = pd.DataFrame({
        "id": [1, 1, 1],
        "metabolite": ["A", "B", "C"]
    })
    
    pd.testing.assert_frame_equal(result.reset_index(drop=True), expected)

def test_explode_empty_dataframe():
    """If the input DataFrame is empty, result should be empty."""
    df = pd.DataFrame(columns=["id", "metabolite"])
    result = build_database.explode_ids(df, "metabolite")
    assert result.empty

def test_explode_single_value():
    """If the column has a single value with no separator, it should remain unchanged."""
    df = pd.DataFrame({"id": [1], "metabolite": ["X"]})
    result = build_database.explode_ids(df, "metabolite")
    expected = pd.DataFrame({"id": [1], "metabolite": ["X"]})
    pd.testing.assert_frame_equal(result.reset_index(drop=True), expected)


def test_explode_with_duplicates():
    """Duplicate values should be preserved unless explicitly removed."""
    df = pd.DataFrame({"id": [1], "metabolite": ["A|A|B"]})
    result = build_database.explode_ids(df, "metabolite")
    expected = pd.DataFrame({
        "id": [1, 1, 1],
        "metabolite": ["A", "A", "B"]
    })
    pd.testing.assert_frame_equal(result.reset_index(drop=True), expected)


# Simply test 
@pytest.mark.parametrize("func,val,expected", [
    # BiGG
    (build_database.simplify_bigg, "bigg.metabolite:glc__D", "glc__D"),
    (build_database.simplify_bigg, "biggM:M_glc__D", "glc__D"),
    (build_database.simplify_bigg, "M_atp|biggM:M_adp|biggM:Pi", "Pi|adp|atp"),
    (build_database.simplify_bigg, None, ""),
    
    # SEED
    (build_database.simplify_seed, "seed.compound:cpd00001", "cpd00001"),
    (build_database.simplify_seed, "M_cpd00002|seedM:cpd00003", "cpd00002|cpd00003"),
    (build_database.simplify_seed, "", ""),
    
    # VMH
    (build_database.simplify_vmh, "vmhM:M_oh1|vmhM:oh1|vmhmetabolite:oh1", "oh1"),
    (build_database.simplify_vmh, "M_lac_L", "lac_L"),
    (build_database.simplify_vmh, None, ""),
    
    # HMDB
    (build_database.simplify_hmdb, "hmdb:HMDB0000123", "HMDB0000123"),
    (build_database.simplify_hmdb, "HMDB0000123|HMDB0000456", "HMDB0000123"),
    (build_database.simplify_hmdb, "foo|bar", "bar"),
    (build_database.simplify_hmdb, None, ""),
    
    # MetaCyc
    (build_database.simplify_metacyc, "metacyc.compound:GLC|metacycM:ATP", "ATP|GLC"),
    (build_database.simplify_metacyc, "GLC", "GLC"),
    (build_database.simplify_metacyc, "", ""),
    
    # Remove M_
    (build_database.remove_prefix_M, "M_glc__D|M_atp", "glc__D|atp"),
    (build_database.remove_prefix_M, "", ""),
    (build_database.remove_prefix_M, None, None),
])
def test_simplify_functions(func, val, expected):
    """Generic test for all simplify_* functions."""
    result = func(val)
    assert result == expected


#!/bin/python
# MISTIC Project INRIA/INRAE
# Author Muller Coralie
# Date: 2024/11/10
# Update: 2025/08/-

####################################
#      Build database              #
####################################

# Script to create a .tsv file with the UNIQUE-ID of Metacyc or MetaNetX database
# and the corresponding informations and ID from other database to do
# the link between them The INCHI have been removed because it seems
# not unique for each metabolite in the database.

# Exemple:
#  UNIQUE-ID	     | CHEBI	|   COMMON-NAME	                                        | ABBREV-NAME |	SYNONYMS	 | ADD-COMPLEMENT  | SEED	  |BIGG	HMDB  |	METANETX	| LIGAND-CPD |	REFMET | PUBCHEM |	CAS |	INCHI-KEY  | SMILES  # noqa: E501
# Teichoic-P-Gro-Glc |	132356	| [(2-Glc)-Gro-P]n-Gro-P-ManNAc-GlcNAc-PP-undecaprenol	|             |[syn1, syn2 ] |		           | cpd28247 |			  | MNXM12982	|			 |	       |	     |      |              | CC(C)=CCCC(\C)=C/CCC(\C)=C/CCC(/C)=C/CCC(/C)=C/CCC(/C)=C/CCC(/C)=C/CCC(/C)=C/CCC(/C)=C/CCC(/C)=C/CCC(/C)=C/COP(=O)([O-])OP([O-])(=O)O[C@@H]1([C@@H]([C@H]([C@@H]([C@H](O1)CO)O[C@@H]3(O[C@H](CO)[C@@H](OP([O-])(=O)OC[C@H](O)COP([O-])(=O)OC[C@H](O[C@H]2(O[C@H](CO)[C@@H](O)[C@H](O)[C@@H](O)2))CO[R])[C@H](O)[C@H](NC(=O)C)3))O)NC(C)=O)  # noqa: E501

import argparse
import csv
import json
import logging
import os
import pandas as pd
import re
import sys
import time
from pathlib import Path
import urllib.request

from metanetmap import utils

# set logger
logger = logging.getLogger(__name__)

####################################
#      Utils functions             #
####################################


def split_line(line, column_name, dictionary_temp):
    """Splits a line to extract specific information and adds it to a dictionary.

    Args:
        line (str): A line from the MetaCyc file being read.
        column_name (str): The key under which the extracted
          information will be stored in the dictionary.
        dictionary_temp (dict): The temporary dictionary to update.
    """
    line_name = line.decode("utf-8")  # Convert the binary line to a UTF-8 string
    extract_name = str(line_name).split('"')[1]  # Extract the value between quotes
    dictionary_temp[column_name] = (
        extract_name  # Store the extracted value in the dictionary
    )


def replace_header_line(line, column_name, dictionary_temp):
    """
    Extracts and stores specific information from a header line in the
    MetaCyc file.

    Args:
        line (str): A line from the MetaCyc file being read (in binary format).
        column_name (str): The key under which the extracted
          information will be stored in the dictionary.
        dictionary_temp (dict): The temporary dictionary to update.
    """
    line_name = line.decode("utf-8")  # Convert the binary line to a UTF-8 string
    extract_name = (
        str(line_name).replace(f"{column_name} - ", "").replace("\n", "")
    )  # Clean and extract the value
    dictionary_temp[column_name] = (
        extract_name  # Store the extracted value in the dictionary
    )


def replace_line(line):
    """
    Cleans a line from a MetaCyc file by removing unwanted prefixes, HTML tags,
    and other characters to prepare it for insertion into a dictionary.

    Specifically designed for COMMON-NAME, SYNONYMS, and ABBREV-NAME lines.

    Args:
        line (bytes): The raw line read from the MetaCyc file (in bytes).

    Returns:
        line.strip() (str): Cleaned and simplified string.
    """
    # Decode from bytes to string
    line = line.decode("utf-8")

    # Remove known prefixes
    prefixes = ["COMMON-NAME - ", "SYNONYMS - ", "ABBREV-NAME - "]
    for prefix in prefixes:
        if line.startswith(prefix):
            line = line[len(prefix) :]
            break  # Only one prefix should match

    # Replace common HTML entities and symbols
    replacements = {"&": "", ";": "", "rarr": "->", "\n": "", "a ": "", "an ": ""}
    for old, new in replacements.items():
        line = line.replace(old, new)

    # Remove HTML tags (e.g., <i>, </i>, <sup>, <SUB>, etc.)
    line = re.sub(r"</?(i|I|sup|SUP|sub|SUB)>", "", line)

    return line.strip()



###########################################
#          Main Script Metacyc            #
###########################################


def build_main_dictionary(metacyc_file):
    """
    Builds a main dictionary from a MetaCyc file using the 'UNIQUE-ID'
    as the primary key, and maps various identifiers from other
    databases for each metabolite.

        Exemple of the structure :
        {'UNIQUE-ID': 'CPD-17659', 'CHEBI': '85306', 'COMMON-NAME':
          'D-allo-isoleucine', 'ABBREV-NAME': '', 'SYNONYMS': '',
          'ADD-COMPLEMENT': '', 'SEED': '', 'BIGG': '', 'HMDB': '',
          'METANETX': 'MNXM17053', 'LIGAND-CPD': 'C21092', 'REFMET':
          '', 'PUBCHEM': '6950184', 'CAS': '', 'INCHI-KEY':
          'InChIKey=AGPKZVBTJJNPAG-CRCLSJGQSA-N', 'SMILES':
          'CC[C@H](C)[C@@H]([NH3+])C([O-])=O'}, {...}

    Args:
        metacyc_file (path): Path to the MetaCyc flat file (usually in
          plain-text or .dat format).

    Returns:
        dictionary_db (list of dict): A list of dictionaries, each
          representing one metabolite, with keys for each relevant field.
        keys (list): A list of all the column names/keys used in the
          dictionary, useful for creating a TSV or DataFrame.
    """
    # 1. Try opening and reading the file.
    # Define all possible keys
    keys = [
        "UNIQUE-ID",
        "CHEBI",
        "COMMON-NAME",
        "ABBREV-NAME",
        "SYNONYMS",
        "ADD-COMPLEMENT",
        "MOLECULAR-WEIGHT",
        "MONOISOTOPIC-MW",
        "SEED",
        "BIGG",
        "HMDB",
        "METANETX",
        "LIGAND-CPD",
        "REFMET",
        "PUBCHEM",
        "CAS",
        "INCHI",
        "NON-STANDARD-INCHI",
        "INCHI-KEY",
        "SMILES",
    ]
    optional_fields = keys[1:]  # All except UNIQUE-ID

    dictionary_db = []
    dictionary_temp = {}
    list_synonyms = []

    try:
        with open(metacyc_file, "rb") as file:
            for line in file:

                if b"UNIQUE-ID" in line:
                    # Append previous metabolite before starting new one
                    if dictionary_temp:
                        # Finalize synonyms as string or empty string if none
                        if list_synonyms:
                            dictionary_temp["SYNONYMS"] = list_synonyms
                        else:
                            dictionary_temp["SYNONYMS"] = ""

                        dictionary_db.append(dictionary_temp)
                        dictionary_temp = {}
                        list_synonyms = []

                    # Extract UNIQUE-ID and initialize new dict
                    line_id = line.decode("utf-8").strip()
                    unique_id = line_id.replace("UNIQUE-ID - ", "").strip()
                    dictionary_temp["UNIQUE-ID"] = unique_id

                    # Initialize all optional fields as empty
                    for field in optional_fields:
                        dictionary_temp[field] = ""

                elif b"CHEBI" in line:
                    split_line(line, "CHEBI", dictionary_temp)

                elif b"COMMON-NAME" in line:
                    dictionary_temp["COMMON-NAME"] = replace_line(line)

                elif b"SYNONYMS" in line:
                    synonym = replace_line(line)
                    list_synonyms.append(synonym)

                elif b"ABBREV-NAME" in line:
                    dictionary_temp["ABBREV-NAME"] = replace_line(line)

                elif b"MOLECULAR-WEIGHT" in line:
                    replace_header_line(line, "MOLECULAR-WEIGHT", dictionary_temp)

                elif b"MONOISOTOPIC-MW" in line:
                    replace_header_line(line, "MONOISOTOPIC-MW", dictionary_temp)

                elif b"SEED" in line:
                    split_line(line, "SEED", dictionary_temp)

                elif b"BIGG" in line:
                    split_line(line, "BIGG", dictionary_temp)

                elif b'(HMDB "' in line:
                    split_line(line, "HMDB", dictionary_temp)

                elif b"METANETX " in line:
                    split_line(line, "METANETX", dictionary_temp)

                elif b"LIGAND-CPD " in line:
                    split_line(line, "LIGAND-CPD", dictionary_temp)

                elif b"REFMET " in line:
                    split_line(line, "REFMET", dictionary_temp)

                elif b"PUBCHEM " in line:
                    split_line(line, "PUBCHEM", dictionary_temp)

                elif b"CAS " in line:
                    split_line(line, "CAS", dictionary_temp)

                elif b"INCHI - " in line and b"NON-STANDARD" not in line:
                    replace_header_line(line, "INCHI", dictionary_temp)

                elif b"NON-STANDARD-INCHI - " in line:
                    replace_header_line(line, "NON-STANDARD-INCHI", dictionary_temp)

                elif b"INCHI-KEY - " in line:
                    replace_header_line(line, "INCHI-KEY", dictionary_temp)

                elif b"SMILES" in line:
                    replace_header_line(line, "SMILES", dictionary_temp)

            # After finishing reading the file, append the last metabolite
            if dictionary_temp:
                if list_synonyms:
                    dictionary_temp["SYNONYMS"] = list_synonyms
                else:
                    dictionary_temp["SYNONYMS"] = ""
                dictionary_db.append(dictionary_temp)

        return dictionary_db, keys

    except Exception as e:
        logger.critical(f"Failed to read MetaCyc file '{metacyc_file}': {e}")
        return [], keys


def manage_synonyms(dictionary_db):
    """
    Processes the 'SYNONYMS' field in each dictionary entry by
    converting lists of synonyms into a JSON-formatted string (e.g.,
    '["syn1", "syn2"]') to ensure compatibility with TSV/CSV export.

    Args:
        dictionary_db (list of dict): The main dictionary list
        containing metabolite data, where each entry may have a
        'SYNONYMS' field as a list.

    Returns:
        dictionary_db (list of dict): The updated dictionary list with
        'SYNONYMS' fields converted to JSON strings (if they were lists).
    """
    for dic in dictionary_db:
        for key, value in dic.items():
            if key == "SYNONYMS":
                if isinstance(value, list):
                    s_with_double_quotes = json.dumps(value)
                    dic[key] = s_with_double_quotes

    return dictionary_db



##############################################
#          Add complement                    #
##############################################


def load_database_metanetx(database_conversion):
    """
    Load the database_conversion load_database_metanetx as a dictionary
    to add complement

    Exemple of the structure :
          [{'UNIQUE-ID': 'CPD-17659', 'CHEBI': '85306', 'COMMON-NAME':
          'D-allo-isoleucine', 'ABBREV-NAME': '',
          'SYNONYMS': [], 'ADD-COMPLEMENT': '', 'SEED': '', 'BIGG':
          '', 'HMDB': '', 'METANETX': 'MNXM17053', 'LIGAND-CPD':
          'C21092',
          'REFMET': '', 'PUBCHEM': '6950184', 'CAS': '', 'INCHI-KEY':
          'InChIKey=AGPKZVBTJJNPAG-CRCLSJGQSA-N', 'SMILES':
          'CC[C@H](C)[C@@H]([NH3+])C([O-])=O'},{...}]

    Args:
        database_conversion (path): Path to the database_conversion to
        exhange into dict

    Returns:
        datatable (list): List with all the dictionnaries with
        database_conversion information -> one dic corresponding to
        one row
    """
    datatable = []
    path = Path(database_conversion)

    # Verify that the database_conversion file has been properly loaded
    try:
        with path.open(encoding="utf-8") as f:
            reader = csv.DictReader(f, delimiter="\t")
            for row in reader:
                datatable.append(row)
    except Exception as e:
        logger.critical(f"Error reading file: {e}")
        sys.exit(1)
    return datatable

def load_complementary_datatable(datatable_conversion):
    """
    Loads a complementary TSV file where the first column must be
    'UNIQUE-ID'.  All other columns are dynamically extracted and
    stored as individual dictionaries mapping UNIQUE-ID to their
    respective values.

    Args:
        datatable_conversion (str): Path to the complementary TSV file.

    Returns:
        dict[str, dict]: A dictionary of dictionaries, each keyed by
        column name (except 'UNIQUE-ID'), mapping UNIQUE-ID to the
        corresponding value.
          Example:
          {
              'BIGG': {'CPD-123': 'glc', ...},
              'SEED': {'CPD-123': 'cpd00027', ...},
              ...
          }

    Raises:
        FileNotFoundError: If the file does not exist.
        ValueError: If 'UNIQUE-ID' is missing.
        IOError: On read failure.
    """
    if not os.path.isfile(datatable_conversion):
        logger.critical(f"Complementary file not found: {datatable_conversion}")
        raise FileNotFoundError(f"File not found: {datatable_conversion}")

    try:
        with open(datatable_conversion, newline="") as csv_f:
            reader = csv.DictReader(csv_f, delimiter="\t")

            if "UNIQUE-ID" not in reader.fieldnames:
                raise ValueError("Missing 'UNIQUE-ID' column in complementary file.")

            # Initialise a dict of dicts for each column (except UNIQUE-ID)
            column_dicts = {col: {} for col in reader.fieldnames if col != "UNIQUE-ID"}

            for row in reader:
                uid = row["UNIQUE-ID"]
                for col in column_dicts:
                    val = row.get(col, "").strip()
                    if val:  # Ignore empty values
                        column_dicts[col][uid] = val

        logger.info(
            f"Loaded complementary file '{datatable_conversion}' with "
            f"columns: {list(column_dicts.keys())}"
        )
        return column_dicts

    except Exception as e:
        logger.critical(f"Failed to load complementary datatable: {e}")
        raise IOError(f"Error reading complementary file: {e}")


def add_each_complement(dictionary_to_add, column_name_add, dictionary_db):
    """Adds complementary information to a specific column in the main dictionary.

    Args:
        dictionary_to_add (dict): Mapping from UNIQUE-ID to a value
          (for one specific column).
        column_name_add (str): Name of the column to add.
        dictionary_db (list of dict): The main list of metabolite dictionaries.

    Returns:
        dictionary_db(list of dict): Updated dictionary_db with the
          new column added where applicable. If a UNIQUE-ID is in
          dictionary_to_add but not in dictionary_db, a new entry will
          be added.
    """
    # Create a lookup for fast matching
    existing_ids = {d["UNIQUE-ID"]: d for d in dictionary_db if "UNIQUE-ID" in d}

    for uid, value in dictionary_to_add.items():
        if uid in existing_ids:
            existing_ids[uid][column_name_add] = value
        else:
            # Add new entry if UID not found
            dictionary_db.append({"UNIQUE-ID": uid, column_name_add: value})

    return dictionary_db


# /!\ WARNING /!\
# The complementary file need two columns at least:
# -> col1: UNIQUE-ID corresponding to the metabolites id in MetaCyc
#    file where we want more information,
# -> col2 the column with information to add


# Exemple: UNIQUE-ID | ADD-COMPLEMENT
#   CPD-4211 | N(2)-succinyl-L-arginine
def add_complement_from_complementary(dictionary_db, complementary_dicts):
    """
    Adds all complementary columns from a complementary file to the main dictionary.

    Args:
        dictionary_db (list of dict): Main metabolite list from the MetaCyc file.
        complementary_dicts (dict): Output from load_complementary_datatable(),
                                    with one sub-dictionary per column to add.

    Returns:
        list of dict: The updated dictionary with all complementary columns added.
    """
    for column_name, mapping in complementary_dicts.items():
        dictionary_db = add_each_complement(mapping, column_name, dictionary_db)
    return dictionary_db



###########################################
#          Main Script MetaNetX           #
###########################################



def download_metanetx_file(filename, url, root) :
    """
    Download a MetaNetX file into a choosen directory and return its full path.

    If the download fails or the file does not exist after download, returns None.

    Parameters:
    -----------
    filename (str) : The name to save the downloaded file as (e.g., 'chem_prop.tsv.gz').
    url (str) : The full URL to download the file from.

    Returns:
    --------
    str or None
        Full path to the downloaded file if successful, None otherwise.
    """
    output = os.path.join(root, "metanetx_db")
    os.makedirs(output, exist_ok=True)

    filepath = os.path.join(output, filename)
    logger.info(f"Downloading {filename} from {url}...")

    try:
        urllib.request.urlretrieve(url, filepath)
        logger.info(f"{filename} successfully downloaded to {filepath}")
    except Exception as e:
        logger.info(f"Error downloading {filename}: {e}")
        return None

    # Check if file exists after download
    if os.path.exists(filepath):
        logger.info(f"File exists: {filepath}")
        return filepath
    else:
        logger.info(f"File not found after download: {filepath}")
        return None




###  File reading functions 
def read_chem_prop(path):
    """
    Read the MetaNetX chemical properties file (chem_prop.tsv).

    Parameters
    ----------
    path (str) Path to the chem_prop.tsv file.

    Returns
    -------
    pd.DataFrame
        DataFrame containing compound properties such as MNX_ID, COMMON_NAME, FORMULA, etc.
    """
    rows = []
    with open(path, encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            parts = line.split("\t")
            rows.append({
                "MNX_ID": parts[0] if len(parts) > 0 else "",
                "COMMON_NAME": parts[1] if len(parts) > 1 else "",
                "REFERENCE": parts[2] if len(parts) > 2 else "",
                "FORMULA": parts[3] if len(parts) > 3 else "",
                "CHARGE": parts[4] if len(parts) > 4 else "",
                "MOLECULAR_WEIGHT": parts[5] if len(parts) > 5 else "",
                "INCHI": parts[6] if len(parts) > 6 else "",
                "INCHI_KEY": parts[7] if len(parts) > 7 else "",
                "SMILES": parts[8] if len(parts) > 8 else ""
            })
    return pd.DataFrame(rows)



def read_chem_xref(path):
    """
    Read the MetaNetX cross-reference file (chem_xref.tsv).

    Parameters
    ----------
    path (str): Path to the chem_xref.tsv file.

    Returns
    -------
    pd.DataFrame
        DataFrame containing cross-reference information between MetaNetX IDs and other databases.
    """
    rows = []
    with open(path, encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            parts = line.split("\t")
            rows.append({
                "source": parts[0] if len(parts) > 0 else "",
                "MNX_ID": parts[1] if len(parts) > 1 else "",
                "description": parts[2] if len(parts) > 2 else ""
            })
    return pd.DataFrame(rows)



# --- Extract database-specific IDs ---
def extract_ids(df, db_prefix):
    """
    Extract all external IDs for a given database prefix.

    Parameters
    ----------
    df : pd.DataFrame
        Cross-reference dataframe.
    db_prefix : str
        Prefix identifying the external database (e.g. 'chebi', 'seed', 'hmdb').

    Returns
    -------
    pd.DataFrame
        A dataframe containing MNX_ID and the concatenated list of unique database-specific identifiers.
    """
    subset = df[df["source"].str.startswith(db_prefix)]
    return subset.groupby("MNX_ID")["source"].apply(lambda x: "|".join(sorted(set(x)))).reset_index().rename(columns={"source": db_prefix.upper()})



def explode_ids(df, col):
    """
    Split a '|' separated column into multiple rows.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    col : str
        Column name containing '|' separated values.

    Returns
    -------
    pd.DataFrame
        A new dataframe with one row per unique ID value.
    """
    exploded = []
    for _, row in df.iterrows():
        for val in row[col].split("|"):
            val = val.strip()
            if val:
                exploded.append({**row, col: val})
    return pd.DataFrame(exploded)


def simplify_bigg(val):
    """
    Clean and simplify BiGG metabolite identifiers.

    Removes known prefixes like 'bigg.metabolite:', 'biggM:', or 'M_'.
    """
    if pd.isna(val): return ""
    vals = set()
    for v in val.split("|"):
        v = re.sub(r"^(bigg\.metabolite:|biggM:M_|biggM:|M_)", "", v.strip())
        if v: vals.add(v)
    return "|".join(sorted(vals))


def simplify_seed(val):
    """
    Clean and simplify SEED compound identifiers.

    Removes prefixes such as 'seed.compound:', 'seedM:', and 'M_'.
    """
    if pd.isna(val): return ""
    vals = set()
    for v in val.split("|"):
        v = re.sub(r"^(seed\.compound:|seedM:|M_)", "", v.strip())
        if v: vals.add(v)
    return "|".join(sorted(vals))

def simplify_vmh(val):
    """
    Clean and simplify VMH compound identifiers.

    Removes prefixes such as 'vmhM:', 'vmhM:M_', 'vmhmetabolite:', and 'M_'.
    Ensures that the final identifier is clean (e.g. 'vmhM:M_oh1|vmhM:oh1|vmhmetabolite:oh1' -> 'oh1').
    """
    if pd.isna(val): return ""
    vals = set()
    for v in val.split("|"):
        # Remove all known VMH prefixes
        v = re.sub(r"^(vmhM:M_|vmhM:|vmhmetabolite:|M_)", "", v.strip())
        if v: vals.add(v)
    return "|".join(sorted(vals))

def simplify_hmdb(val):
    """
    Clean and normalize HMDB identifiers.

    If multiple IDs are found, the function keeps the first valid HMDB####### entry.
    """
    if pd.isna(val) or val.strip() == "":
        return ""
    vals = [v.replace("hmdb:", "").strip() for v in val.split("|")]
    valid = [v for v in vals if re.match(r"^HMDB\d{7}$", v)]
    if valid:
        return sorted(valid)[0]
    return sorted(set(vals))[0] if vals else ""


def simplify_metacyc(val):
    """
    Clean and simplify MetaCyc compound identifiers.

    Removes prefixes such as 'metacyc.compound:' and 'metacycM:'.
    """
    if pd.isna(val) or val.strip() == "":
        return ""
    vals = set()
    for v in val.split("|"):
        v = re.sub(r"^(metacyc\.compound:|metacycM:)", "", v.strip())
        if v: vals.add(v)
    return "|".join(sorted(vals))

# --- Final cleanup: remove any remaining 'M_' prefix in SEED and BIGG columns ---
def remove_prefix_M(val):
    """
    Remove the prefix 'M_' from all non-empty cells in a given string.

    Handles multiple IDs separated by '|'.
    """
    if pd.isna(val) or val.strip() == "":
        return val
    vals = [re.sub(r"^M_", "", v.strip()) for v in val.split("|")]
    return "|".join(vals)



######################
#       Run          #
######################


def parse_args():
    """
    Parses command-line arguments for the MetaCyc/MetaNetX data conversion script.

    Returns:
        argparse.Namespace: Parsed arguments 
    """
    parser = argparse.ArgumentParser()

    return parser.parse_args()


def load_args(args=None):
    """
    Main function to run the MetaCyc/MetaNetX to TSV conversion pipeline.

    Args:
        args (argparse.Namespace, optional): Parsed command-line arguments.
            If None, the arguments will be parsed from sys.argv.

    This function:
        - Parses the MetaCyc .dat or TSV from MetaNetX files to extract compound information.
        - Optionally loads a complementary file for additional
          identifiers (BIGG, SEED, etc.).
        - Builds the conversion dictionary.
        - Writes the output to a TSV file.
        - Runs in quiet mode if specified (suppresses logs except for warnings/errors).
    """
    if args is None:
        args = parse_args()

    # -------------------------------#
    #       Set up the logger       #
    # -------------------------------#

    t0 = time.time()

    logger.setLevel(logging.DEBUG)

    # ----Clean the root logger (ROOT) ----
    root_logger = logging.getLogger()
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    # Create leveled formatter
    formatter = utils.LeveledFormatter("%(message)s")
    formatter.set_formatter(logging.INFO, logging.Formatter("%(message)s"))
    formatter.set_formatter(
        logging.WARNING,
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"),
    )
    formatter.set_formatter(
        logging.CRITICAL,
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"),
    )

    # Set up the default console logger
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    if args.quiet:
        console_handler.setLevel(logging.WARNING)
        print("---->    Build database run in quiet mode    <----\n")
    logger.addHandler(console_handler)


    if not args.output:
        output = ""
    else:
        output = args.output

    # ----------------------------------------#
    #        Check validity of output         #
    # ----------------------------------------#

    # Manage the creation of the output file and directory if they are not
    utils.is_valid_file_or_create(output)
    root, ext = os.path.splitext(output)
    if not ext:  # Check if we have the path for outputfolder but not the name
        # in args, create an arbitrary filename.
        utils.is_valid_dir(root)
        if args.db == "metacyc":
            utils.is_valid_file_or_create(f"{root}metacyc_conversion_datatable.tsv")
            output = os.path.join(root, "metacyc_conversion_datatable.tsv")
        elif args.db == "metanetx":
            utils.is_valid_file_or_create(f"{root}metanetx_conversion_datatable.tsv")
            output = os.path.join(root, "metanetx_conversion_datatable.tsv")
        else:
            logger.critical(
            "Error: the '--db' argument must be either 'metacyc' or 'metanetx'. "
            "You must choose one of these database methods."
            )
            sys.exit()
    else:
        stripped = os.path.dirname(output)
        utils.is_valid_dir(stripped)

    

    directory = os.path.dirname(output)
    if directory != "":
        utils.is_valid_dir(f"{directory}/logs/")
        output_log = f"{directory}/logs/"
    else:
        utils.is_valid_dir("logs/")
        output_log = "logs/"

    # #-----------------------------------#
    # #       Set up the logger file      #
    # #------------------------------------#
    # #  # # File handler
    log_file_path = os.path.join(output_log, f"build_database_{t0}.log")
    file_handler = logging.FileHandler(log_file_path, "w+")
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    # ---------------------------------------#
    #      Set up complementary file        #
    # ---------------------------------------#

    logger.info("-----------------------------------------")
    logger.info("            BUILD CONVERTION TABLE ")
    logger.info("----------------------------------------- \n")

    # Check if we have a complementary file : The complementary file
    # allow to add information that are not in the metacyc file.
    complementary_file = args.complement_file

    # ----------------------#
    #    Main command      #
    # ----------------------#

    logger.info("\nCommand run:")
    logger.info("Actual command run (from sys.argv): python " + " ".join(sys.argv))

    if args.db == "metacyc":
        utils.is_valid_file(args.metacyc_file)
        if complementary_file:  # iF complementrary file is add
            utils.is_valid_file(complementary_file)

            logger.info("\n---> Run construction of the datatable for Metacyc\n")
            dictionary_db, keys = build_main_dictionary(args.metacyc_file)
            dictionary_db = manage_synonyms(dictionary_db)

            logger.info("---> Complementary file added :")
            column_dicts = load_complementary_datatable(args.complement_file)
            dictionary_db = add_complement_from_complementary(dictionary_db, column_dicts)
            utils.write_csv(dictionary_db, output, keys)
        else:
            logger.info("\n---> Run construction of the datatable for Metacyc")
            logger.info("/!\\  No complementary file added")

            dictionary_db, keys = build_main_dictionary(args.metacyc_file)
            dictionary_db = manage_synonyms(dictionary_db)
            utils.write_csv(dictionary_db, output, keys)
    
    if args.db == "metanetx":
        if args.chem_prop_file == "":
            chem_prop_file=download_metanetx_file("chem_prop.tsv","https://www.metanetx.org/cgi-bin/mnxget/mnxref/chem_prop.tsv",root)
        else:
            chem_prop_file=args.chem_prop_file
        if args.chem_ref_file =="":   
            chem_ref_file=download_metanetx_file("chem_xref.tsv","https://www.metanetx.org/cgi-bin/mnxget/mnxref/chem_xref.tsv",root)
        else:
            chem_ref_file=args.chem_ref_file
        if complementary_file:  # iF complementrary file is add
            utils.is_valid_file(complementary_file)
    
            logger.info("\n---> Run construction of the datatable for MetaNetX\n")
                        # --- Load data ---
            df_prop = read_chem_prop(chem_prop_file)
            df_xref = read_chem_xref(chem_ref_file)
            # Extract references for each database
            chebi_df = extract_ids(df_xref, "chebi")
            pubchem_df = extract_ids(df_xref, "pubchem")
            bigg_df = extract_ids(df_xref, "bigg")
            hmdb_df = extract_ids(df_xref, "hmdb")
            refmet_df = extract_ids(df_xref, "refmet")
            seed_df = extract_ids(df_xref, "seed")
            vmh_df = extract_ids(df_xref, "vmh")  

            # --- Extract MetaCyc references ---
            metacyc_df = df_xref[df_xref["source"].str.startswith("metacyc")][["MNX_ID", "source"]].rename(columns={"source": "METACYC"})

            # --- Specific cleaning functions ---
            # CHEBI → one row per ID
            chebi_df = explode_ids(chebi_df, "CHEBI")

            bigg_df["BIGG"] = bigg_df["BIGG"].apply(simplify_bigg)
            bigg_df = explode_ids(bigg_df, "BIGG")
            seed_df["SEED"] = seed_df["SEED"].apply(simplify_seed)
            seed_df = explode_ids(seed_df, "SEED")

            vmh_df["VMH"] = vmh_df["VMH"].apply(simplify_vmh)
            vmh_df = explode_ids(vmh_df, "VMH")

            hmdb_df["HMDB"] = hmdb_df["HMDB"].apply(simplify_hmdb)
            metacyc_df["METACYC"] = metacyc_df["METACYC"].apply(simplify_metacyc)

            # --- Merge all dataframes together ---
            df_final = df_prop.merge(metacyc_df, on="MNX_ID", how="left")
            for d in [chebi_df, pubchem_df, bigg_df, hmdb_df, refmet_df, seed_df, vmh_df]:
                df_final = df_final.merge(d, on="MNX_ID", how="left")

            # Rename MNX_ID → UNIQUE-ID
            df_final = df_final.rename(columns={"MNX_ID": "UNIQUE-ID"})

            # Remove CHARGE column if present
            if "CHARGE" in df_final.columns:
                df_final = df_final.drop(columns=["CHARGE"])

            # Add missing columns if necessary
            for col in ["ABBREV_NAME", "SYNONYMS", "ADD-COMPLEMENT", "NON-STANDARD-INCHI"]:
                if col not in df_final.columns:
                    df_final[col] = ""

            # Reorder columns for final export (added VMH)
            cols_order = [
                "UNIQUE-ID", "CHEBI", "COMMON_NAME", "ABBREV_NAME", "SYNONYMS", "ADD-COMPLEMENT",
                "MOLECULAR_WEIGHT", "SEED", "BIGG", "HMDB", "METACYC",
                "REFMET", "PUBCHEM", "VMH", "CAS", "INCHI", "NON-STANDARD-INCHI", "INCHI_KEY", "SMILES"
            ]
            for c in cols_order:
                if c not in df_final.columns:
                    df_final[c] = ""

            df_final = df_final[cols_order]

            df_final["SEED"] = df_final["SEED"].apply(remove_prefix_M)
            df_final["BIGG"] = df_final["BIGG"].apply(remove_prefix_M)

            # --- Remove full duplicates ---
            df_final = df_final.drop_duplicates()

            # --- Save final merged table ---
            df_final.to_csv(output, sep="\t", index=False)
            logger.info(f"Final merged table generated: {output}")
    
            logger.info("---> Complementary file added :")
            column_dicts = load_complementary_datatable(args.complement_file)
            dictionary_db_metanetx=load_database_metanetx(output)
            dictionary_db = add_complement_from_complementary(dictionary_db_metanetx, column_dicts)
            utils.write_csv(dictionary_db, output, cols_order)


        else:
            logger.info("\n---> Run construction of the datatable for MetaNetX")
            logger.info("/!\\  No complementary file added")
            
            # --- Load data ---
            df_prop = read_chem_prop(chem_prop_file)
            df_xref = read_chem_xref(chem_ref_file)
            # Extract references for each database
            chebi_df = extract_ids(df_xref, "chebi")
            pubchem_df = extract_ids(df_xref, "pubchem")
            bigg_df = extract_ids(df_xref, "bigg")
            hmdb_df = extract_ids(df_xref, "hmdb")
            refmet_df = extract_ids(df_xref, "refmet")
            seed_df = extract_ids(df_xref, "seed")
            vmh_df = extract_ids(df_xref, "vmh")  

            # --- Extract MetaCyc references ---
            metacyc_df = df_xref[df_xref["source"].str.startswith("metacyc")][["MNX_ID", "source"]].rename(columns={"source": "METACYC"})

            # --- Specific cleaning functions ---
            # CHEBI → one row per ID
            chebi_df = explode_ids(chebi_df, "CHEBI")

            bigg_df["BIGG"] = bigg_df["BIGG"].apply(simplify_bigg)
            bigg_df = explode_ids(bigg_df, "BIGG")
            seed_df["SEED"] = seed_df["SEED"].apply(simplify_seed)
            seed_df = explode_ids(seed_df, "SEED")

            vmh_df["VMH"] = vmh_df["VMH"].apply(simplify_vmh)
            vmh_df = explode_ids(vmh_df, "VMH")

            hmdb_df["HMDB"] = hmdb_df["HMDB"].apply(simplify_hmdb)
            metacyc_df["METACYC"] = metacyc_df["METACYC"].apply(simplify_metacyc)

            # --- Merge all dataframes together ---
            df_final = df_prop.merge(metacyc_df, on="MNX_ID", how="left")
            for d in [chebi_df, pubchem_df, bigg_df, hmdb_df, refmet_df, seed_df, vmh_df]:
                df_final = df_final.merge(d, on="MNX_ID", how="left")

            # Rename MNX_ID → UNIQUE-ID
            df_final = df_final.rename(columns={"MNX_ID": "UNIQUE-ID"})

            # Remove CHARGE column if present
            if "CHARGE" in df_final.columns:
                df_final = df_final.drop(columns=["CHARGE"])

            # Add missing columns if necessary
            for col in ["ABBREV_NAME", "SYNONYMS", "ADD-COMPLEMENT", "NON-STANDARD-INCHI"]:
                if col not in df_final.columns:
                    df_final[col] = ""

            # Reorder columns for final export (added VMH)
            cols_order = [
                "UNIQUE-ID", "CHEBI", "COMMON_NAME", "ABBREV_NAME", "SYNONYMS", "ADD-COMPLEMENT",
                "MOLECULAR_WEIGHT", "SEED", "BIGG", "HMDB", "METACYC",
                "REFMET", "PUBCHEM", "VMH", "CAS", "INCHI", "NON-STANDARD-INCHI", "INCHI_KEY", "SMILES"
            ]
            for c in cols_order:
                if c not in df_final.columns:
                    df_final[c] = ""

            df_final = df_final[cols_order]

            df_final["SEED"] = df_final["SEED"].apply(remove_prefix_M)
            df_final["BIGG"] = df_final["BIGG"].apply(remove_prefix_M)

            # --- Remove full duplicates ---
            df_final = df_final.drop_duplicates()

            # --- Save final merged table ---
            df_final.to_csv(output, sep="\t", index=False)
            logger.info(f"Final merged table generated: {output}")

    t1 = time.time()
    if args.quiet:
        print(
            "\n--- Total runtime %.2f seconds ---"
            "\n ---> Construction of the database completed" % (t1 - t0)
        )
        print("")
    else:
        logger.info(
            "\n--- Total runtime %.2f seconds ---"
            "\n ---> Construction of the database completed" % (t1 - t0)
        )


if __name__ == "__main__":
    load_args()
